# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Band-descriptors for blocked/strided groups.

This module contains classes defining two kinds of band groups:

* Blocked groups with contiguous band indices.
* Strided groups with evenly-spaced band indices.
"""

import numpy as np

from gpaw.mpi import serial_comm

NONBLOCKING = False


class BandDescriptor:
    r"""Descriptor-class for ordered lists of bands

    A ``BandDescriptor`` object holds information on how functions, such
    as wave functions and corresponding occupation numbers, are divided
    into groups according to band indices. The main information here is
    how many bands are stored on each processor and who gets what.

    This is how a 12 band array is laid out in memory on 3 cpu's::

      a) Blocked groups       b) Strided groups

           3    7   11             9   10   11
     myn   2 \  6 \ 10       myn   6    7    8
      |    1  \ 5  \ 9        |    3    4    5
      |    0    4    8        |    0    1    2
      |                       |
      +----- band_rank        +----- band_rank

    Example:

     >>> a = np.zeros((3, 4))
     >>> a.ravel()[:] = range(12)
     >>> a
     array([[ 0.,  1.,  2.,  3.],
            [ 4.,  5.,  6.,  7.],
            [ 8.,  9., 10., 11.]])
     >>> b = np.zeros((4, 3))
     >>> b.ravel()[:] = range(12)
     >>> b.T
     array([[ 0.,  3.,  6.,  9.],
            [ 1.,  4.,  7., 10.],
            [ 2.,  5.,  8., 11.]])
     """

    def __init__(self, nbands: int, comm=None, strided=False):
        """Construct band-descriptor object.

        Parameters:

        nbands: int
            Global number of bands.
        comm: MPI-communicator
            Communicator for band-groups.
        strided: bool
            Enable strided band distribution for better
            load balancing with many unoccupied bands.

        Note that if comm.size is 1, then all bands are contained on
        a single CPU and blocked/strided grouping loses its meaning.

        Attributes:

        ============  ======================================================
        ``nbands``    Number of bands in total.
        ``mynbands``  Number of bands on this CPU.
        ``beg``       Beginning of band indices in group (inclusive).
        ``end``       End of band indices in group (exclusive).
        ``step``      Stride for band indices between ``beg`` and ``end``.
        ``comm``      MPI-communicator for band distribution.
        ============  ======================================================
        """

        if comm is None:
            comm = serial_comm

        self.comm = comm
        self.nbands = nbands
        self.strided = strided

        self.maxmynbands = (nbands + comm.size - 1) // comm.size

        if strided:
            assert nbands % comm.size == 0
            self.beg = comm.rank
            self.end = nbands
            self.step = comm.size
            self.mynbands = self.maxmynbands
        else:
            self.beg = min(nbands, comm.rank * self.maxmynbands)
            self.end = min(nbands, self.beg + self.maxmynbands)
            self.mynbands = self.end - self.beg
            self.step = 1

    def __len__(self):
        return self.mynbands

    def get_slice(self, band_rank=None):
        """Return the slice of global bands which belong to a given rank."""
        if band_rank is None:
            band_rank = self.comm.rank
        assert band_rank in range(self.comm.size)

        if self.strided:
            nstride = self.comm.size
            nslice = slice(band_rank, None, nstride)
        else:
            nslice = slice(self.beg, self.end)
        return nslice

    def get_band_indices(self, band_rank=None):
        """Return the global band indices which belong to a given rank."""
        nslice = self.get_slice(band_rank)
        return np.arange(*nslice.indices(self.nbands))

    def get_band_ranks(self):
        """Return array of ranks as a function of global band indices."""
        rank_n = np.empty(self.nbands, dtype=int)
        for band_rank in range(self.comm.size):
            nslice = self.get_slice(band_rank)
            rank_n[nslice] = band_rank
        assert (rank_n >= 0).all() and (rank_n < self.comm.size).all()
        return rank_n

    def who_has(self, n):
        """Convert global band index to rank information and local index."""
        if self.strided:
            myn, band_rank = divmod(n, self.comm.size)
        else:
            band_rank, myn = divmod(n, self.maxmynbands)
        return band_rank, myn

    def global_index(self, myn, band_rank=None):
        """Convert rank information and local index to global index."""
        if band_rank is None:
            band_rank = self.comm.rank

        if self.strided:
            return band_rank + myn * self.comm.size

        return band_rank * self.maxmynbands + myn

    def get_size_of_global_array(self):
        return (self.nbands,)

    def zeros(self, n=(), dtype=float, global_array=False):
        """Return new zeroed 3D array for this domain.

        The type can be set with the ``dtype`` keyword (default:
        ``float``).  Extra dimensions can be added with ``n=dim``.
        A global array spanning all domains can be allocated with
        ``global_array=True``."""
        # TODO XXX doc
        return self._new_array(n, dtype, True, global_array)

    def empty(self, n=(), dtype=float, global_array=False):
        """Return new uninitialized 3D array for this domain.

        The type can be set with the ``dtype`` keyword (default:
        ``float``).  Extra dimensions can be added with ``n=dim``.
        A global array spanning all domains can be allocated with
        ``global_array=True``."""
        # TODO XXX doc
        return self._new_array(n, dtype, False, global_array)

    def _new_array(self, n=(), dtype=float, zero=True, global_array=False):
        if global_array:
            shape = self.get_size_of_global_array()
        else:
            shape = (self.mynbands,)

        if isinstance(n, int):
            n = (n,)

        shape = tuple(shape) + n

        if zero:
            return np.zeros(shape, dtype)
        else:
            return np.empty(shape, dtype)

    def collect(self, a_nx, broadcast=False):
        """Collect distributed array to master-CPU or all CPU's."""
        if self.comm.size == 1:
            return a_nx

        xshape = a_nx.shape[1:]

        # Optimization for blocked groups
        if not self.strided:
            if self.comm.size * self.maxmynbands > self.nbands:
                return self.nasty_non_strided_collect(a_nx, broadcast)
            if broadcast:
                A_nx = self.empty(xshape, a_nx.dtype, global_array=True)
                self.comm.all_gather(a_nx, A_nx)
                return A_nx

            if self.comm.rank == 0:
                A_nx = self.empty(xshape, a_nx.dtype, global_array=True)
            else:
                A_nx = None
            self.comm.gather(a_nx, 0, A_nx)
            return A_nx

        # Collect all arrays on the master:
        if self.comm.rank != 0:
            # There can be several sends before the corresponding receives
            # are posted, so use syncronous send here
            self.comm.ssend(a_nx, 0, 3011)
            if broadcast:
                A_nx = self.empty(xshape, a_nx.dtype, global_array=True)
                self.comm.broadcast(A_nx, 0)
                return A_nx
            else:
                return None

        # Put the band groups from the slaves into the big array
        # for the whole collection of bands:
        A_nx = self.empty(xshape, a_nx.dtype, global_array=True)
        for band_rank in range(self.comm.size):
            if band_rank != 0:
                a_nx = self.empty(xshape, a_nx.dtype, global_array=False)
                self.comm.receive(a_nx, band_rank, 3011)
            A_nx[self.get_slice(band_rank), ...] = a_nx

        if broadcast:
            self.comm.broadcast(A_nx, 0)
        return A_nx

    def distribute(self, B_nx, b_nx):
        """ distribute full array B_nx to band groups, result in
        b_nx. b_nx must be allocated."""

        S = self.comm.size

        if S == 1:
            b_nx[:] = B_nx
            return

        # Optimization for blocked groups
        if not self.strided:
            M2 = self.maxmynbands
            if M2 * S == self.nbands:
                self.comm.scatter(B_nx, b_nx, 0)
                return

            if self.comm.rank == 0:
                C_nx = np.empty((S * M2,) + B_nx.shape[1:], B_nx.dtype)
                C_nx[:self.nbands] = B_nx
            else:
                C_nx = None

            if self.mynbands < M2:
                c_nx = np.empty((M2,) + b_nx.shape[1:], b_nx.dtype)
            else:
                c_nx = b_nx

            self.comm.scatter(C_nx, c_nx, 0)

            if self.mynbands < M2:
                b_nx[:] = c_nx[:self.mynbands]

            return

        if self.comm.rank != 0:
            self.comm.receive(b_nx, 0, 421)
            return
        else:
            requests = []
            for band_rank in range(self.comm.size):
                if band_rank != 0:
                    a_nx = B_nx[self.get_slice(band_rank), ...].copy()
                    request = self.comm.send(a_nx, band_rank, 421, NONBLOCKING)
                    # Remember to store a reference to the
                    # send buffer (a_nx) so that is isn't
                    # deallocated:
                    requests.append((request, a_nx))
                else:
                    b_nx[:] = B_nx[self.get_slice(), ...]

            for request, a_nx in requests:
                self.comm.wait(request)

    def nasty_non_strided_collect(self, a_nx, broadcast):
        xshape = a_nx.shape[1:]
        if broadcast:
            A_nx = self.nasty_non_strided_collect(a_nx, False)
            if A_nx is None:
                A_nx = self.empty(xshape, a_nx.dtype, global_array=True)
            self.comm.broadcast(A_nx, 0)
            return A_nx

        S = self.comm.size
        M2 = self.maxmynbands
        if self.comm.rank == 0:
            A_nx = np.empty((S * M2,) + xshape, a_nx.dtype)
        else:
            A_nx = None

        if self.mynbands < M2:
            b_nx = np.empty((M2,) + xshape, a_nx.dtype)
            b_nx[:self.mynbands] = a_nx
        else:
            b_nx = a_nx

        self.comm.gather(b_nx, 0, A_nx)

        if self.comm.rank > 0:
            return

        return A_nx[:self.nbands]



"""
Calculate the magnetic response in iron using ALDA.

Fast test, where the kernel is scaled to fulfill the Goldstone theorem.
"""

# Workflow modules
from pathlib import Path
import pytest
import numpy as np

# Script modules
from gpaw.test import findpeak
from gpaw.mpi import world

from gpaw.response import ResponseGroundStateAdapter, ResponseContext
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.susceptibility import (ChiFactory, spectral_decomposition,
                                          read_eigenmode_lineshapes)
from gpaw.response.localft import LocalGridFTCalculator, LocalPAWFTCalculator
from gpaw.response.fxc_kernels import FXCKernel, AdiabaticFXCCalculator
from gpaw.response.dyson import HXCScaling, HXCKernel
from gpaw.response.pair_functions import read_pair_function


def set_up_fxc_calculators(gs, context):
    fxckwargs_and_identifiers = []

    # Set up grid calculator (without file buffer)
    localft_calc = LocalGridFTCalculator(gs, context)
    fxckwargs_grid = {'fxc_calculator': AdiabaticFXCCalculator(localft_calc),
                      'hxc_scaling': HXCScaling('fm')}
    fxckwargs_and_identifiers.append((fxckwargs_grid, 'grid'))

    # Set up paw calculator (with file buffer)
    localft_calc = LocalPAWFTCalculator(gs, context, rshelmax=0)
    fxckwargs_paw = {'fxc_calculator': AdiabaticFXCCalculator(localft_calc),
                     'fxc_file': Path('paw_ALDA_fxc.npz'),
                     'hxc_scaling': HXCScaling('fm')}
    fxckwargs_and_identifiers.append((fxckwargs_paw, 'paw'))

    return fxckwargs_and_identifiers


def get_test_values(identifier):
    test_mw1 = 0.  # meV
    test_Ipeak1 = 7.48  # a.u.
    test_Ipeak3 = 19.83  # a.u.
    if identifier == 'grid':
        test_fxcs = 1.059
        test_mw2 = 363.  # meV
        test_Ipeak2 = 3.47  # a.u.
        test_Ipeak4 = 9.20  # a.u.
    elif identifier == 'paw':
        test_fxcs = 1.131
        test_mw2 = 352.  # meV
        test_Ipeak2 = 3.35  # a.u.
        test_Ipeak4 = 9.14  # a.u.
    else:
        raise ValueError(f'Invalid identifier {identifier}')

    return (test_fxcs, test_mw1, test_mw2,
            test_Ipeak1, test_Ipeak2, test_Ipeak3, test_Ipeak4)


@pytest.mark.kspair
@pytest.mark.response
def test_response_iron_sf_gssALDA(in_tmp_dir, gpw_files):
    # ---------- Inputs ---------- #

    nbands = 6
    q_qc = [[0.0, 0.0, 0.0], [0.0, 0.0, 1. / 4.]]  # Two q-points along G-N
    frq_qw = [np.linspace(-0.080, 0.120, 26), np.linspace(0.250, 0.450, 26)]
    fxc = 'ALDA'
    ecut = 300
    eta = 0.1
    if world.size > 1:
        nblocks = 2
    else:
        nblocks = 1

    # ---------- Script ---------- #

    context = ResponseContext()
    gs = ResponseGroundStateAdapter.from_gpw_file(gpw_files['fe_pw'],
                                                  context=context)

    chiks_calc = ChiKSCalculator(gs,
                                 context=context,
                                 nbands=nbands,
                                 ecut=ecut,
                                 gammacentered=True,
                                 nblocks=nblocks)
    fxckwargs_and_identifiers = set_up_fxc_calculators(gs, context)

    chi_factory = ChiFactory(
        chiks_calc,
        # Use the first fxc_calculator for the ChiFactory
        fxc_calculator=fxckwargs_and_identifiers[0][0]['fxc_calculator'])

    for q in range(2):
        complex_frequencies = frq_qw[q] + 1.j * eta

        # Calculate chi using the various fxc calculators
        for f, (fxckwargs, identifier) in enumerate(fxckwargs_and_identifiers):
            if f == 0:
                # The first kernel is used directly through the ChiFactory
                chiks, chi = chi_factory('+-', q_qc[q], complex_frequencies,
                                         fxc=fxc,
                                         hxc_scaling=fxckwargs['hxc_scaling'])
            else:  # We wish to test storing the remaining kernels in files
                assert 'fxc_file' in fxckwargs
                fxc_file = fxckwargs['fxc_file']
                if q == 0:  # Calculate and save kernel
                    assert not fxc_file.is_file()
                    fxc_calculator = fxckwargs['fxc_calculator']
                    fxc_kernel = fxc_calculator(fxc, '+-', chiks.qpd)
                    fxc_kernel.save(fxc_file)
                else:  # Reuse kernel from previous calculation
                    assert fxc_file.is_file()
                    fxc_kernel = FXCKernel.from_file(fxc_file)

                # Calculate many-body susceptibility
                hxc_kernel = HXCKernel(None, fxc_kernel,
                                       fxckwargs['hxc_scaling'])
                chi = chi_factory.dyson_solver(chiks, hxc_kernel)

            chi.write_macroscopic_component(identifier + '_iron_dsus'
                                            + '_%d.csv' % (q + 1))
            Amaj, _ = spectral_decomposition(chi, pos_eigs=10, neg_eigs=0)
            Amaj.write_eigenmode_lineshapes(identifier + '_Amaj_modes'
                                            + '_%d.csv' % (q + 1), nmodes=1)

        chi_factory.context.write_timer()

    world.barrier()

    # plot_comparison('grid', 'paw')

    # Compare results to test values
    for fxckwargs, identifier in fxckwargs_and_identifiers:
        fxcs = fxckwargs['hxc_scaling'].lambd
        (_, _, mw1, Ipeak1, _, _, mw2, Ipeak2,
         _, _, mw3, Ipeak3, _, _, mw4, Ipeak4) = extract_data(identifier)

        print(fxcs, mw1, mw2, Ipeak1, Ipeak2, mw3, Ipeak3, mw4, Ipeak4)

        (test_fxcs, test_mw1, test_mw2,
         test_Ipeak1, test_Ipeak2,
         test_Ipeak3, test_Ipeak4) = get_test_values(identifier)

        # fxc scaling:
        assert fxcs == pytest.approx(test_fxcs, abs=0.005)

        # Magnon peak:
        assert mw1 == pytest.approx(test_mw1, abs=20.)
        assert mw2 == pytest.approx(test_mw2, abs=50.)
        assert mw3 == pytest.approx(mw1, abs=10.)
        assert mw4 == pytest.approx(mw2, abs=10.)

        # Scattering function intensity:
        assert Ipeak1 == pytest.approx(test_Ipeak1, abs=0.5)
        assert Ipeak2 == pytest.approx(test_Ipeak2, abs=0.5)
        assert Ipeak3 == pytest.approx(test_Ipeak3, abs=0.5)
        assert Ipeak4 == pytest.approx(test_Ipeak4, abs=1.1)


def extract_data(identifier):
    # Read data
    w1_w, chi1_w = read_pair_function(identifier + '_iron_dsus_1.csv')
    w2_w, chi2_w = read_pair_function(identifier + '_iron_dsus_2.csv')
    w3_w, s3_wm = read_eigenmode_lineshapes(identifier + '_Amaj_modes_1.csv')
    w4_w, s4_wm = read_eigenmode_lineshapes(identifier + '_Amaj_modes_2.csv')
    s3_w = s3_wm[:, 0]
    s4_w = s4_wm[:, 0]

    # Spectral function
    S1_w = -chi1_w.imag
    S2_w = -chi2_w.imag

    # Identify peaks
    wpeak1, Ipeak1 = findpeak(w1_w, S1_w)
    wpeak2, Ipeak2 = findpeak(w2_w, S2_w)
    wpeak3, Ipeak3 = findpeak(w3_w, s3_w)
    wpeak4, Ipeak4 = findpeak(w4_w, s4_w)

    # Peak positions in meV
    mw1 = wpeak1 * 1000
    mw2 = wpeak2 * 1000
    mw3 = wpeak3 * 1000
    mw4 = wpeak4 * 1000

    return (w1_w, S1_w, mw1, Ipeak1, w2_w, S2_w, mw2, Ipeak2,
            w3_w, s3_w, mw3, Ipeak3, w4_w, s4_w, mw4, Ipeak4)


def plot_comparison(identifier1, identifier2):
    (w11_w, S11_w, _, _, w12_w, S12_w, _, _,
     w13_w, s13_w, _, _, w14_w, s14_w, _, _) = extract_data(identifier1)
    (w21_w, S21_w, _, _, w22_w, S22_w, _, _,
     w23_w, s23_w, _, _, w24_w, s24_w, _, _) = extract_data(identifier2)

    import matplotlib.pyplot as plt
    plt.subplot(2, 2, 1)
    plt.plot(w11_w, S11_w)
    plt.plot(w21_w, S21_w)
    plt.subplot(2, 2, 2)
    plt.plot(w12_w, S12_w)
    plt.plot(w22_w, S22_w)
    plt.subplot(2, 2, 3)
    plt.plot(w13_w, s13_w)
    plt.plot(w23_w, s23_w)
    plt.subplot(2, 2, 4)
    plt.plot(w14_w, s14_w)
    plt.plot(w24_w, s24_w)
    plt.show()



import pytest
from ase.build import graphene_nanoribbon, molecule
from gpaw import GPAW
from gpaw.lcao.local_orbitals import LocalOrbitals
from gpaw.mpi import world

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


@pytest.mark.later
def test_gamma_point_calculation():
    atoms = molecule('C6H6', vacuum=2)

    calc = GPAW(mode='lcao',
                xc='LDA',
                basis='szp(dzp)',
                txt=None)
    atoms.calc = calc
    atoms.get_potential_energy()

    los = LocalOrbitals(calc)

    los.subdiagonalize('C', groupby='energy')
    group_energies = sorted(los.groups.keys())

    # Assert minimal model contains only pz-LOs
    los.take_model(minimal=True)
    assert los.indices == los.groups[group_energies[3]]
    assert len(los.indices) == 6

    # Assert extended model also contains +2 d-LOs
    los.take_model(minimal=False)
    assert los.indices == (
        los.groups[group_energies[3]]
        + los.groups[group_energies[-3]]
        + los.groups[group_energies[-1]])
    assert len(los.indices) == (6 * 3)


@pytest.mark.later
def test_k_point_calculation():
    atoms = graphene_nanoribbon(2, 1, type='zigzag', saturated=True,
                                C_H=1.1, C_C=1.4, vacuum=2)

    calc = GPAW(mode='lcao',
                xc='LDA',
                basis='szp(dzp)',
                txt=None,
                kpts={'size': (1, 1, 11),
                      'gamma': True},
                symmetry={'point_group': False,
                          'time_reversal': True})
    atoms.calc = calc
    atoms.get_potential_energy()

    los = LocalOrbitals(calc)

    los.subdiagonalize('C', groupby='symmetry')
    group_energies = sorted(los.groups.keys())

    # Assert minimal model contains only pz-LOs
    los.take_model(minimal=True)
    assert los.indices == los.groups[group_energies[3]]
    assert len(los.indices) == 4

    # Assert extended model also contains d-LOs
    los.take_model(minimal=False)
    assert los.indices == (
        los.groups[group_energies[3]]
        + los.groups[group_energies[-3]]
        + los.groups[group_energies[-1]])
    assert len(los.indices) == (4 * 3)



import numpy as np

from gpaw.mpi import serial_comm
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.lfc import PWLFC
from gpaw.hybrids.kpts import PWKPoint, RSKPoint, to_real_space
from gpaw.utilities.blas import mmm


def apply1(kpt, Htpsit_xG, wfs, coulomb, sym, paw):
    kd = wfs.kd
    kpts = [PWKPoint(kpt.psit,
                     kpt.projections,
                     kpt.f_n / kpt.weight,  # scale to [0, 1] range
                     kd.ibzk_kc[kpt.k],
                     kd.weight_k[kpt.k])
            for kpt in wfs.kpt_u[kpt.s::wfs.nspins]]
    evv, evc, ekin, v_knG = calculate(kpts, wfs, paw, sym, coulomb)
    return evc, evv, ekin, v_knG


def calculate(kpts, wfs, paw, sym, coulomb):
    pd = kpts[0].psit.pd
    gd = pd.gd.new_descriptor(comm=serial_comm)
    kd = wfs.kd
    comm = wfs.world
    nbands = len(kpts[0].psit.array)
    shapes = [(nbands, len(Delta_iiL))
              for Delta_iiL in paw.Delta_aiiL]
    v_kani = [{a: np.zeros(shape, pd.dtype)
               for a, shape in enumerate(shapes)}
              for _ in range(len(kpts))]
    v_knG = [k.psit.pd.zeros(nbands, global_array=True, q=k.psit.kpt)
             for k in kpts]

    exxvv = 0.0
    ekin = 0.0
    for i1, i2, s, k1, k2, count in sym.pairs(kpts, wfs, wfs.spos_ac):
        q_c = k2.k_c - k1.k_c
        qd = KPointDescriptor([-q_c])

        pd12 = PWDescriptor(pd.ecut, gd, pd.dtype, kd=qd)
        ghat = PWLFC([data.ghat_l for data in wfs.setups], pd12)
        ghat.set_positions(wfs.spos_ac)

        v1_nG = v_knG[i1]
        v1_ani = v_kani[i1]
        v2_nG = v_knG[i2]
        v2_ani = v_kani[i2]

        v_G = coulomb.get_potential(pd12)
        assert i1 == kpts[i1].psit.kpt
        assert i2 == kpts[i2].psit.kpt
        e_nn = calculate_exx_for_pair(k1, k2, ghat, v_G,
                                      kpts[i1].psit.pd,
                                      kpts[i2].psit.pd,
                                      kpts[i1].psit.kpt,
                                      kpts[i2].psit.kpt,
                                      k1.f_n,
                                      k2.f_n,
                                      s,
                                      count,
                                      v1_nG, v1_ani,
                                      v2_nG, v2_ani,
                                      wfs, sym, paw)

        e_nn *= count
        e = k1.f_n.dot(e_nn).dot(k2.f_n) / kd.nbzkpts
        exxvv -= 0.5 * e
        ekin += e

    exxvc = 0.0
    for i, kpt in enumerate(kpts):
        for a, VV_ii in paw.VV_aii.items():
            P_ni = kpt.proj[a]
            vv_n = np.einsum('ni, ij, nj -> n',
                             P_ni.conj(), VV_ii, P_ni).real
            vc_n = np.einsum('ni, ij, nj -> n',
                             P_ni.conj(), paw.VC_aii[a], P_ni).real
            exxvv -= vv_n.dot(kpt.f_n) * kpt.weight
            exxvc -= vc_n.dot(kpt.f_n) * kpt.weight

    w_knG = {}
    G1 = comm.rank * pd.maxmyng
    G2 = (comm.rank + 1) * pd.maxmyng
    for v_nG, v_ani, kpt in zip(v_knG, v_kani, kpts):
        comm.sum(v_nG)
        w_nG = v_nG[:, G1:G2].copy()
        w_knG[len(w_knG)] = w_nG
        for v_ni in v_ani.values():
            comm.sum(v_ni)
        v1_ani = {}
        for a, VV_ii in paw.VV_aii.items():
            P_ni = kpt.proj[a]
            v_ni = P_ni.dot(paw.VC_aii[a] + 2 * VV_ii)
            v1_ani[a] = v_ani[a] - v_ni
            ekin += (np.einsum('n, ni, ni',
                               kpt.f_n, P_ni.conj(), v_ni).real *
                     kpt.weight)
        wfs.pt.add(w_nG, v1_ani, kpt.psit.kpt)

    return comm.sum(exxvv), comm.sum(exxvc), comm.sum(ekin), w_knG


def calculate_exx_for_pair(k1,
                           k2,
                           ghat,
                           v_G,
                           pd1, pd2,
                           index1, index2,
                           f1_n, f2_n,
                           s,
                           count,
                           v1_nG,
                           v1_ani,
                           v2_nG,
                           v2_ani,
                           wfs,
                           sym,
                           paw,
                           F_av=None):
    kd = wfs.kd
    comm = wfs.world
    factor = 1.0 / kd.nbzkpts

    N1 = len(k1.u_nR)
    N2 = len(k2.u_nR)

    size = comm.size
    rank = comm.rank

    Q_annL = [np.einsum('mi, ijL, nj -> mnL',
                        k1.proj[a],
                        Delta_iiL,
                        k2.proj[a].conj(),
                        optimize=True)
              for a, Delta_iiL in enumerate(paw.Delta_aiiL)]

    if v2_nG is not None:
        T, T_a, cc = sym.symmetry_operation(s, wfs, inverse=True)

    if k1 is k2:
        n2max = (N1 + size - 1) // size
    else:
        n2max = N2

    e_nn = np.zeros((N1, N2))
    rho_nG = ghat.pd.empty(n2max, k1.u_nR.dtype)
    vrho_nG = ghat.pd.empty(n2max, k1.u_nR.dtype)

    f_GI = ghat.expand()

    for n1, u1_R in enumerate(k1.u_nR):
        if k1 is k2:
            B = (N1 - n1 + size - 1) // size
            n20 = 0
            n2a = min(n1 + rank * B, N2)
            n2b = min(n2a + B, N2)
        else:
            B = (N1 + size - 1) // size
            n20 = min(B * rank, N1)
            n2a = 0
            n2b = N2

        for n2, rho_G in enumerate(rho_nG[:n2b - n2a], n2a):
            rho_G[:] = ghat.pd.fft(u1_R * k2.u_nR[n2].conj())

        add(ghat, rho_nG[:n2b - n2a],
            {a: Q_nnL[n1, n2a:n2b]
             for a, Q_nnL in enumerate(Q_annL)},
            f_GI)

        for n2, rho_G in enumerate(rho_nG[:n2b - n2a], n2a):
            vrho_G = v_G * rho_G
            if F_av:
                for a, v_xL in ghat.derivative(vrho_G).items():
                    print(a, v_xL.shape)
                1 / 0
            e = ghat.pd.integrate(rho_G, vrho_G).real
            e_nn[n1, n2] = e
            if k1 is k2:
                e_nn[n2, n1] = e
            vrho_nG[n2 - n2a] = vrho_G

            if v1_nG is not None:
                vrho_R = ghat.pd.ifft(vrho_G)
                if v2_nG is None:
                    assert k1 is not k2
                    v1_nG[n1] -= f2_n[n2] * factor * pd1.fft(
                        vrho_R * k2.u_nR[n2], index1, local=True)
                else:
                    x = factor * count / 2
                    if k1 is k2 and n1 != n2:
                        x *= 2
                    x1 = x / (kd.weight_k[index1] * kd.nbzkpts)
                    x2 = x / (kd.weight_k[index2] * kd.nbzkpts)
                    v1_nG[n1] -= f2_n[n2] * x1 * pd1.fft(
                        vrho_R * k2.u_nR[n2], index1, local=True)
                    v2_nG[n2 + n20] -= f1_n[n1] * x2 * pd2.fft(
                        T(vrho_R.conj() * u1_R), index2,
                        local=True)

        if v1_nG is not None and v2_nG is None:
            for a, v_nL in integrate(ghat, vrho_nG[:n2b - n2a], f_GI):
                v_iin = paw.Delta_aiiL[a].dot(v_nL.T)
                v1_ani[a][n1] -= np.einsum('ijn, nj, n -> i',
                                           v_iin,
                                           k2.proj[a][n2a:n2b],
                                           f2_n[n2a:n2b] * factor)

        if v1_nG is not None and v2_nG is not None:
            x = factor * count / kd.nbzkpts / 2
            x1 = x / kd.weight_k[index1]
            x2 = x / kd.weight_k[index2]
            if k1 is k2:
                x1 *= 2
                x2 *= 2

            for a, v_nL in integrate(ghat, vrho_nG[:n2b - n2a], f_GI):
                if k1 is k2 and n2a <= n1 < n2b:
                    v_nL[n1 - n2a] *= 0.5
                v_iin = paw.Delta_aiiL[a].dot(v_nL.T)
                v1_ani[a][n1] -= np.einsum('ijn, nj, n -> i',
                                           v_iin,
                                           k2.proj[a][n2a:n2b],
                                           f2_n[n2a:n2b] * x1)
                b, S_c, U_ii = T_a[a]
                v_ni = np.einsum('ijn, j, ik -> nk',
                                 v_iin.conj(),
                                 k1.proj[b][n1],
                                 U_ii)
                if v_nL.dtype == complex:
                    v_ni *= np.exp(2j * np.pi * k2.k_c.dot(S_c))
                    if cc:
                        v_ni = v_ni.conj()
                v2_ani[a][n20 + n2a:n20 + n2b] -= v_ni * f1_n[n1] * x2

    return e_nn * factor


def add(ghat, a_xG, c_axi, f_GI):
    c_xI = np.empty(a_xG.shape[:-1] + (ghat.nI,), ghat.pd.dtype)
    for a, I1, I2 in ghat.my_indices:
        c_xI[..., I1:I2] = c_axi[a] * ghat.eikR_qa[0][a].conj()
    nx = np.prod(c_xI.shape[:-1], dtype=int)
    c_xI = c_xI.reshape((nx, ghat.nI))
    a_xG = a_xG.reshape((nx, a_xG.shape[-1])).view(ghat.pd.dtype)
    mmm(1.0 / ghat.pd.gd.dv, c_xI, 'N', f_GI, 'T', 1.0, a_xG)


def integrate(ghat, a_xG, f_GI):
    c_xI = np.zeros(a_xG.shape[:-1] + (ghat.nI,), ghat.pd.dtype)

    nx = np.prod(c_xI.shape[:-1], dtype=int)
    b_xI = c_xI.reshape((nx, ghat.nI))
    a_xG = a_xG.reshape((nx, a_xG.shape[-1]))

    alpha = 1.0 / ghat.pd.gd.N_c.prod()
    if ghat.pd.dtype == float:
        alpha *= 2
        a_xG = a_xG.view(float)
        f_GI[0] *= 0.5
    else:
        f_GI.imag[:] = -f_GI.imag
    mmm(alpha, a_xG, 'N', f_GI, 'N', 0.0, b_xI)
    if ghat.pd.dtype == complex:
        f_GI.imag[:] = -f_GI.imag
    else:
        f_GI[0] *= 2.0
    for a, I1, I2 in ghat.my_indices:
        yield a, ghat.eikR_qa[0][a] * c_xI[..., I1:I2]


def apply2(kpt, psit_xG, Htpsit_xG, wfs, coulomb, sym, paw):
    kd = wfs.kd

    psit = kpt.psit.new(buf=psit_xG)
    P = kpt.projections.new()
    psit.matrix_elements(wfs.pt, out=P)

    kpt1 = PWKPoint(psit,
                    P,
                    kpt.f_n + np.nan,
                    kd.ibzk_kc[kpt.k],
                    np.nan)

    kpts2 = [PWKPoint(kpt.psit,
                      kpt.projections,
                      kpt.f_n / kpt.weight,  # scale to [0, 1] range
                      kd.ibzk_kc[kpt.k],
                      kd.weight_k[kpt.k])
             for kpt in wfs.kpt_u[kpt.s::wfs.nspins]]
    v_nG = calculate2(kpt1, kpts2, wfs, paw, sym, coulomb)
    return v_nG


def calculate2(kpt1, kpts2, wfs, paw, sym, coulomb):
    pd = kpt1.psit.pd
    gd = pd.gd.new_descriptor(comm=serial_comm)
    kd = wfs.kd
    comm = wfs.world
    nbands = len(kpt1.psit.array)
    shapes = [(nbands, len(Delta_iiL))
              for Delta_iiL in paw.Delta_aiiL]
    v_ani = {a: np.zeros(shape, pd.dtype)
             for a, shape in enumerate(shapes)}
    v_nG = kpt1.psit.pd.zeros(nbands, global_array=True, q=kpt1.psit.kpt)

    u1_nR = to_real_space(kpt1.psit)
    proj1 = kpt1.proj.broadcast()
    k1 = RSKPoint(u1_nR,
                  proj1,
                  kpt1.f_n,
                  kpt1.k_c,
                  kpt1.weight)

    N2 = len(kpts2[0].psit.array)
    nsym = len(kd.symmetry.op_scc)

    size = comm.size
    rank = comm.rank
    B = (N2 + size - 1) // size
    na = min(B * rank, N2)
    nb = min(na + B, N2)
    for i2, kpt2 in enumerate(kpts2):
        u2_nR = to_real_space(kpt2.psit, na, nb)
        k0 = RSKPoint(u2_nR,
                      kpt2.proj.broadcast().view(na, nb),
                      kpt2.f_n[na:nb],
                      kpt2.k_c,
                      kpt2.weight)
        for k, i in enumerate(kd.bz2ibz_k):
            if i != i2:
                continue
            s = kd.sym_k[k] + kd.time_reversal_k[k] * nsym
            k2 = sym.apply_symmetry(s, k0, wfs, wfs.spos_ac)
            q_c = k2.k_c - k1.k_c
            qd = KPointDescriptor([-q_c])

            pd12 = PWDescriptor(pd.ecut, gd, pd.dtype, kd=qd)
            ghat = PWLFC([data.ghat_l for data in wfs.setups], pd12)
            ghat.set_positions(wfs.spos_ac)

            v_G = coulomb.get_potential(pd12)
            calculate_exx_for_pair(k1, k2, ghat, v_G,
                                   kpt1.psit.pd,
                                   kpts2[i2].psit.pd,
                                   kpt1.psit.kpt,
                                   kpts2[i2].psit.kpt,
                                   k1.f_n,
                                   k2.f_n,
                                   s,
                                   1.0,
                                   v_nG, v_ani,
                                   None, None,
                                   wfs, sym, paw)

    G1 = comm.rank * pd.maxmyng
    G2 = (comm.rank + 1) * pd.maxmyng
    comm.sum(v_nG)
    w_nG = v_nG[:, G1:G2].copy()
    for v_ni in v_ani.values():
        comm.sum(v_ni)
    v1_ani = {}
    for a, VV_ii in paw.VV_aii.items():
        P_ni = kpt1.proj[a]
        v_ni = P_ni.dot(paw.VC_aii[a] + 2 * VV_ii)
        v1_ani[a] = v_ani[a] - v_ni
    wfs.pt.add(w_nG, v1_ani, kpt1.psit.kpt)

    return w_nG



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""K-point descriptor."""

from __future__ import annotations
from typing import Optional, Sequence

import numpy as np
from ase.calculators.calculator import kptdensity2monkhorstpack
from ase.dft.kpoints import get_monkhorst_pack_size_and_offset, monkhorst_pack

import _gpaw
import gpaw.mpi as mpi
from gpaw import KPointError
from gpaw.typing import Array1D
from gpaw.kpoint import KPoint


def to1bz(bzk_kc, cell_cv):
    """Wrap k-points to 1. BZ.

    Return k-points wrapped to the 1. BZ.

    bzk_kc: (n,3) ndarray
        Array of k-points in units of the reciprocal lattice vectors.
    cell_cv: (3,3) ndarray
        Unit cell.
    """

    B_cv = 2.0 * np.pi * np.linalg.inv(cell_cv).T
    K_kv = np.dot(bzk_kc, B_cv)
    N_xc = np.indices((3, 3, 3)).reshape((3, 27)).T - 1
    G_xv = np.dot(N_xc, B_cv)

    bz1k_kc = bzk_kc.copy()

    # Find the closest reciprocal lattice vector:
    for k, K_v in enumerate(K_kv):
        # If a k-point has the same distance to several reciprocal
        # lattice vectors, we don't want to pick a random one on the
        # basis of numerical noise, so we round off the differences
        # between the shortest distances to 6 decimals and chose the
        # one with the lowest index.
        d = ((G_xv - K_v)**2).sum(1)
        x = (d - d.min()).round(6).argmin()
        bz1k_kc[k] -= N_xc[x]

    return bz1k_kc


def kpts2sizeandoffsets(size=None, density=None, gamma=None, even=None,
                        atoms=None):
    """Helper function for selecting k-points.

    Use either size or density.

    size: 3 ints
        Number of k-points.
    density: float
        K-point density in units of k-points per Ang^-1.
    gamma: None or bool
        Should the Gamma-point be included?  Yes / no / don't care:
        True / False / None.
    even: None or bool
        Should the number of k-points be even?  Yes / no / don't care:
        True / False / None.
    atoms: Atoms object
        Needed for calculating k-point density.

    """

    if size is None:
        if density is None:
            size = [1, 1, 1]
        else:
            size = kptdensity2monkhorstpack(atoms, density, even)

    offsets = [0, 0, 0]

    if gamma is not None:
        for i, s in enumerate(size):
            if atoms.pbc[i] and s % 2 != bool(gamma):
                offsets[i] = 0.5 / s

    return size, offsets


class KPointDescriptor:
    """Descriptor-class for k-points."""

    def __init__(self, kpts, nspins: int = 1):
        """Construct descriptor object for kpoint/spin combinations (ks-pair).

        Parameters:

        kpts: None, sequence of 3 ints, or (n,3)-shaped array
            Specification of the k-point grid. None=Gamma, list of
            ints=Monkhorst-Pack, ndarray=user specified.
        nspins: int
            Number of spins.

        Attributes
        ===================  =================================================
        ``N_c``               Number of k-points in the different directions.
        ``nspins``            Number of spins in total.
        ``mynspins``          Number of spins on this CPU.
        ``nibzkpts``          Number of irreducible kpoints in 1st BZ.
        ``mynks``             Number of k-point/spin combinations on this CPU.
        ``gamma``             Boolean indicator for gamma point calculation.
        ``comm``              MPI-communicator for kpoint distribution.
        ``weight_k``          Weights of each k-point
        ``ibzk_kc``           Unknown
        ``ibzk_qc``           Unknown
        ``sym_k``             Unknown
        ``time_reversal_k``   Unknown
        ``bz2ibz_k``          Unknown
        ``ibz2bz_k``          Unknown
        ``bz2bz_ks``          Unknown
        ``symmetry``          Object representing symmetries
        ===================  =================================================
        """

        self.N_c: Optional[Array1D] = None
        self.offset_c: Optional[Array1D] = None

        if kpts is None:
            self.bzk_kc = np.zeros((1, 3))
            self.N_c = np.array((1, 1, 1), dtype=int)
            self.offset_c = np.zeros(3)
        else:
            kpts = np.asarray(kpts)
            if kpts.ndim == 1:
                self.N_c = np.array(kpts, dtype=int)
                self.bzk_kc = monkhorst_pack(self.N_c)
                self.offset_c = np.zeros(3)
            else:
                self.bzk_kc = np.array(kpts, dtype=float)
                try:
                    self.N_c, self.offset_c = \
                        get_monkhorst_pack_size_and_offset(self.bzk_kc)
                except ValueError:
                    pass
        self.nspins = nspins
        self.nbzkpts = len(self.bzk_kc)

        # Gamma-point calculation?
        self.gamma = self.nbzkpts == 1 and not self.bzk_kc.any()

        # Point group and time-reversal symmetry neglected:
        self.weight_k = np.ones(self.nbzkpts) / self.nbzkpts
        self.ibzk_kc = self.bzk_kc.copy()
        self.sym_k = np.zeros(self.nbzkpts, int)
        self.time_reversal_k = np.zeros(self.nbzkpts, bool)
        self.bz2ibz_k = np.arange(self.nbzkpts)
        self.ibz2bz_k = np.arange(self.nbzkpts)
        self.bz2bz_ks = np.arange(self.nbzkpts)[:, np.newaxis]
        self.nibzkpts = self.nbzkpts
        self.refine_info = None
        self.monkhorst = (self.N_c is not None)

        self.set_communicator(mpi.serial_comm)

    def __str__(self):
        s = str(self.symmetry)

        if self.refine_info is not None:
            s += '\n' + str(self.refine_info)

        if -1 in self.bz2bz_ks:
            s += 'Note: your k-points are not as symmetric as your crystal!\n'

        if self.gamma:
            s += '\n1 k-point (Gamma)'
        else:
            s += '\n%d k-points' % self.nbzkpts
            if self.monkhorst:
                s += ': %d x %d x %d Monkhorst-Pack grid' % tuple(self.N_c)
                if self.offset_c.any():
                    s += ' + ['
                    for x in self.offset_c:
                        if x != 0 and abs(round(1 / x) - 1 / x) < 1e-12:
                            s += '1/%d,' % round(1 / x)
                        else:
                            s += '%f,' % x
                    s = s[:-1] + ']'

        s += ('\n%d k-point%s in the irreducible part of the Brillouin zone\n'
              % (self.nibzkpts, ' s'[1:self.nibzkpts]))

        if self.monkhorst:
            w_k = self.weight_k * self.nbzkpts
            assert np.allclose(w_k, w_k.round())
            w_k = w_k.round()

        s += '       k-points in crystal coordinates                weights\n'
        for k in range(self.nibzkpts):
            if k < 10 or k == self.nibzkpts - 1:
                if self.monkhorst:
                    s += ('%4d:   %12.8f  %12.8f  %12.8f     %6d/%d\n' %
                          ((k,) + tuple(self.ibzk_kc[k]) +
                           (w_k[k], self.nbzkpts)))
                else:
                    s += ('%4d:   %12.8f  %12.8f  %12.8f     %12.8f\n' %
                          ((k,) + tuple(self.ibzk_kc[k]) +
                           (self.weight_k[k],)))
            elif k == 10:
                s += '          ...\n'
        return s

    def set_symmetry(self, atoms, symmetry, comm=None):
        """Create symmetry object and construct irreducible Brillouin zone.

        atoms: Atoms object
            Defines atom positions and types and also unit cell and
            boundary conditions.
        symmetry: Symmetry object
            Symmetry object.
        """

        self.symmetry = symmetry

        # XXX we pass the whole atoms object just to complain if its PBCs
        # are not how we like them
        for c, periodic in enumerate(atoms.pbc):
            if not periodic and not np.allclose(self.bzk_kc[:, c], 0.0):
                raise ValueError('K-points can only be used with PBCs!')

        if symmetry.time_reversal or symmetry.point_group:
            (self.ibzk_kc, self.weight_k,
             self.sym_k,
             self.time_reversal_k,
             self.bz2ibz_k,
             self.ibz2bz_k,
             self.bz2bz_ks) = symmetry.reduce(self.bzk_kc, comm)

        # Number of irreducible k-points and k-point/spin combinations.
        self.nibzkpts = len(self.ibzk_kc)

    def set_communicator(self, comm):
        """Set k-point communicator."""

        # Ranks < self.rank0 have mynks0 k-point/spin combinations and
        # ranks >= self.rank0 have mynks0+1 k-point/spin combinations.
        mynk0, x = divmod(self.nibzkpts, comm.size)
        self.rank0 = comm.size - x
        self.comm = comm

        # My number and offset of k-point/spin combinations
        self.mynk = self.get_count()
        self.k0 = self.get_offset()

        self.ibzk_qc = self.ibzk_kc[self.k0:self.k0 + self.mynk]
        self.weight_q = self.weight_k[self.k0:self.k0 + self.mynk]

    def copy(self, comm=mpi.serial_comm):
        """Create a copy with shared symmetry object."""
        kd = KPointDescriptor(self.bzk_kc, self.nspins)
        kd.weight_k = self.weight_k
        kd.ibzk_kc = self.ibzk_kc
        kd.sym_k = self.sym_k
        kd.time_reversal_k = self.time_reversal_k
        kd.bz2ibz_k = self.bz2ibz_k
        kd.ibz2bz_k = self.ibz2bz_k
        kd.bz2bz_ks = self.bz2bz_ks
        kd.symmetry = self.symmetry
        kd.nibzkpts = self.nibzkpts
        kd.set_communicator(comm)
        return kd

    def create_k_points(self, sdisp_cd, collinear):
        """Return a list of KPoints."""

        kpt_qs = []

        for k in range(self.k0, self.k0 + self.mynk):
            q = k - self.k0
            weightk = self.weight_k[k]
            weight = weightk * 2 / self.nspins
            if self.gamma:
                phase_cd = np.ones((3, 2), complex)
            else:
                phase_cd = np.exp(2j * np.pi *
                                  sdisp_cd * self.ibzk_kc[k, :, np.newaxis])
            if collinear:
                spins = range(self.nspins)
            else:
                spins = [None]
                weight *= 0.5
            kpt_qs.append([KPoint(weightk, weight, s, k, q, phase_cd)
                           for s in spins])

        return kpt_qs

    def collect(self, a_ux, broadcast: bool):
        """Collect distributed data to all."""

        xshape = a_ux.shape[1:]
        a_qsx = a_ux.reshape((-1, self.nspins) + xshape)
        if self.comm.rank == 0 or broadcast:
            a_ksx = np.empty((self.nibzkpts, self.nspins) + xshape, a_ux.dtype)

        if self.comm.rank > 0:
            self.comm.send(a_qsx, 0)
        else:
            k1 = self.get_count(0)
            a_ksx[0:k1] = a_qsx
            requests = []
            for rank in range(1, self.comm.size):
                k2 = k1 + self.get_count(rank)
                requests.append(self.comm.receive(a_ksx[k1:k2], rank,
                                                  block=False))
                k1 = k2
            assert k1 == self.nibzkpts
            self.comm.waitall(requests)

        if broadcast:
            self.comm.broadcast(a_ksx, 0)

        if self.comm.rank == 0 or broadcast:
            return a_ksx.transpose((1, 0, 2))

    def transform_wave_function(self, psit_G, k, index_G=None, phase_G=None):
        """Transform wave function from IBZ to BZ.

        k is the index of the desired k-point in the full BZ.
        """

        s = self.sym_k[k]
        time_reversal = self.time_reversal_k[k]
        op_cc = np.linalg.inv(self.symmetry.op_scc[s]).round().astype(int)

        # Identity
        if (np.abs(op_cc - np.eye(3, dtype=int)) < 1e-10).all():
            if time_reversal:
                return psit_G.conj()
            else:
                return psit_G
        # General point group symmetry
        else:
            ik = self.bz2ibz_k[k]
            kibz_c = self.ibzk_kc[ik]
            b_g = np.zeros_like(psit_G)
            kbz_c = np.dot(self.symmetry.op_scc[s], kibz_c)
            if index_G is not None:
                assert index_G.shape == psit_G.shape == phase_G.shape
                _gpaw.symmetrize_with_index(psit_G, b_g, index_G, phase_G)
            else:
                _gpaw.symmetrize_wavefunction(psit_G, b_g, op_cc.copy(),
                                              np.ascontiguousarray(kibz_c),
                                              kbz_c)

            if time_reversal:
                return b_g.conj()
            else:
                return b_g

    def get_transform_wavefunction_index(self, nG, k):
        """Get the "wavefunction transform index".

        This is a permutation of the numbers 1, 2, .. N which
        associates k + q to some k, and where N is the total
        number of grid points as specified by nG which is a
        3D tuple.

        Returns index_G and phase_G which are one-dimensional
        arrays on the grid."""

        s = self.sym_k[k]
        op_cc = np.linalg.inv(self.symmetry.op_scc[s]).round().astype(int)

        # General point group symmetry
        if (np.abs(op_cc - np.eye(3, dtype=int)) < 1e-10).all():
            nG0 = np.prod(nG)
            index_G = np.arange(nG0).reshape(nG)
            phase_G = np.ones(nG)
        else:
            ik = self.bz2ibz_k[k]
            kibz_c = self.ibzk_kc[ik]
            index_G = np.zeros(nG, dtype=int)
            phase_G = np.zeros(nG, dtype=complex)

            kbz_c = np.dot(self.symmetry.op_scc[s], kibz_c)
            _gpaw.symmetrize_return_index(index_G, phase_G, op_cc.copy(),
                                          np.ascontiguousarray(kibz_c),
                                          kbz_c)
        return index_G, phase_G

    def find_k_plus_q(self, q_c, kpts_k: Sequence[int] = None) -> list[int]:
        """Find the indices of k+q for all kpoints in the Brillouin zone.

        In case that k+q is outside the BZ, the k-point inside the BZ
        corresponding to k+q is given.

        Parameters
        ----------
        q_c: np.ndarray
            Coordinates for the q-vector in units of the reciprocal
            lattice vectors.
        kpts_k:
            Restrict search to specified k-points.

        """
        k_x = kpts_k
        if k_x is None:
            return self.find_k_plus_q(q_c, range(self.nbzkpts))

        i_x = []
        for k in k_x:
            kpt_c = self.bzk_kc[k] + q_c
            d_kc = kpt_c - self.bzk_kc
            d_k = abs(d_kc - d_kc.round()).sum(1)
            i = d_k.argmin()
            if d_k[i] > 1e-8:
                raise KPointError('Could not find k+q!')
            i_x.append(i)

        return i_x

    def get_bz_q_points(self, first=False):
        """Return the q=k1-k2. q-mesh is always Gamma-centered."""
        shift_c = 0.5 * ((self.N_c + 1) % 2) / self.N_c
        bzq_qc = monkhorst_pack(self.N_c) + shift_c
        if first:
            return to1bz(bzq_qc, self.symmetry.cell_cv)
        else:
            return bzq_qc

    def get_ibz_q_points(self, bzq_qc, op_scc):
        """Return ibz q points and the corresponding symmetry operations that
        work for k-mesh as well."""

        ibzq_qc_tmp = []
        ibzq_qc_tmp.append(bzq_qc[-1])
        weight_tmp = [0]

        for i, op_cc in enumerate(op_scc):
            if np.abs(op_cc - np.eye(3)).sum() < 1e-8:
                identity_iop = i
                break

        ibzq_q_tmp = {}
        iop_q = {}
        timerev_q = {}
        diff_qc = {}

        for i in range(len(bzq_qc) - 1, -1, -1):  # loop opposite to kpoint
            try:
                ibzk, iop, timerev, diff_c = self.find_ibzkpt(
                    op_scc, ibzq_qc_tmp, bzq_qc[i])
                find = False
                for ii, iop1 in enumerate(self.sym_k):
                    if iop1 == iop and self.time_reversal_k[ii] == timerev:
                        find = True
                        break
                if not find:
                    raise ValueError('cant find k!')

                ibzq_q_tmp[i] = ibzk
                weight_tmp[ibzk] += 1.
                iop_q[i] = iop
                timerev_q[i] = timerev
                diff_qc[i] = diff_c
            except ValueError:
                ibzq_qc_tmp.append(bzq_qc[i])
                weight_tmp.append(1.)
                ibzq_q_tmp[i] = len(ibzq_qc_tmp) - 1
                iop_q[i] = identity_iop
                timerev_q[i] = False
                diff_qc[i] = np.zeros(3)

        # reverse the order.
        nq = len(ibzq_qc_tmp)
        ibzq_qc = np.zeros((nq, 3))
        ibzq_q = np.zeros(len(bzq_qc), dtype=int)
        for i in range(nq):
            ibzq_qc[i] = ibzq_qc_tmp[nq - i - 1]
        for i in range(len(bzq_qc)):
            ibzq_q[i] = nq - ibzq_q_tmp[i] - 1
        self.q_weights = np.array(weight_tmp[::-1]) / len(bzq_qc)
        return ibzq_qc, ibzq_q, iop_q, timerev_q, diff_qc

    def find_ibzkpt(self, symrel, ibzk_kc, bzk_c):
        """Find index in IBZ and related symmetry operations."""
        find = False
        ibzkpt = 0
        iop = 0
        timerev = False

        for sign in (1, -1):
            for ioptmp, op in enumerate(symrel):
                for i, ibzk in enumerate(ibzk_kc):
                    diff_c = bzk_c - sign * np.dot(op, ibzk)
                    if (np.abs(diff_c - diff_c.round()) < 1e-8).all():
                        ibzkpt = i
                        iop = ioptmp
                        find = True
                        if sign == -1:
                            timerev = True
                        break
                if find:
                    break
            if find:
                break

        if not find:
            raise ValueError('Cant find corresponding IBZ kpoint!')
        return ibzkpt, iop, timerev, diff_c.round()

    def where_is_q(self, q_c, bzq_qc):
        """Find the index of q points in BZ."""
        d_qc = q_c - bzq_qc
        d_q = abs(d_qc - d_qc.round()).sum(1)
        q = d_q.argmin()
        if d_q[q] > 1e-8:
            raise KPointError('Could not find q!')
        return q

    def get_count(self, rank=None):
        """Return the number of ks-pairs which belong to a given rank."""

        if rank is None:
            rank = self.comm.rank
        assert rank in range(self.comm.size)
        mynk0 = self.nibzkpts // self.comm.size
        mynk = mynk0
        if rank >= self.rank0:
            mynk += 1
        return mynk

    def get_offset(self, rank=None):
        """Return the offset of the first ks-pair on a given rank."""

        if rank is None:
            rank = self.comm.rank
        assert rank in range(self.comm.size)
        mynk0 = self.nibzkpts // self.comm.size
        k0 = rank * mynk0
        if rank >= self.rank0:
            k0 += rank - self.rank0
        return k0

    def get_rank_and_index(self, k):
        """Find rank and local index of k-point/spin combination."""

        rank, q = self.who_has(k)
        return rank, q

    def get_indices(self, rank=None):
        """Return the global ks-pair indices which belong to a given rank."""

        k1 = self.get_offset(rank)
        k2 = k1 + self.get_count(rank)
        return np.arange(k1, k2)

    def who_has(self, k):
        """Convert global index to rank information and local index."""

        mynk0 = self.nibzkpts // self.comm.size
        if k < mynk0 * self.rank0:
            rank, q = divmod(k, mynk0)
        else:
            rank, q = divmod(k - mynk0 * self.rank0, mynk0 + 1)
            rank += self.rank0
        return rank, q

    def write(self, writer):
        writer.write('ibzkpts', self.ibzk_kc)
        writer.write('bzkpts', self.bzk_kc)
        writer.write('bz2ibz', self.bz2ibz_k)
        writer.write('weights', self.weight_k)
        writer.write('rotations', self.symmetry.op_scc)
        writer.write('translations', self.symmetry.ft_sc)
        writer.write('atommap', self.symmetry.a_sa)



import numpy as np
from gpaw.utilities.blas import axpy, r2k, rk, gemmdot, mmm, mmmx
from gpaw.utilities.tools import tri2full


def test_gemm_size_zero():
    c = np.ones((3, 3))
    a = np.zeros((0, 3))
    b = np.zeros((3, 0))
    d = np.zeros((0, 0))
    e = np.zeros((0, 3))
    # gemm(1.0, a, b, 0.0, c, 'n')
    mmm(1.0, b, 'N', a, 'N', 0.0, c)
    assert (c == 0.0).all()
    mmm(1.0, d, 'N', a, 'N', 0.0, e)


def test_linalg_blas():
    a = np.arange(5 * 7).reshape(5, 7) + 4.
    a2 = np.arange(3 * 7).reshape(3, 7) + 3.
    b = np.arange(7) - 2.

    # Check gemmdot with floats
    assert np.all(np.dot(a, b) == gemmdot(a, b))
    assert np.all(np.dot(a, a2.T) == gemmdot(a, a2, trans='t'))
    assert np.all(np.dot(a, a2.T) == gemmdot(a, a2, trans='c'))
    assert np.dot(b, b) == gemmdot(b, b)

    # Check gemmdot with complex arrays
    a = a * (2 + 1.j)
    a2 = a2 * (-1 + 3.j)
    b = b * (3 - 2.j)
    assert np.all(np.dot(a, b) == gemmdot(a, b))
    assert np.all(np.dot(a, a2.T) == gemmdot(a, a2, trans='t'))
    assert np.all(np.dot(a, a2.T.conj()) == gemmdot(a, a2, trans='c'))
    assert np.dot(b, b) == gemmdot(b, b, trans='n')
    assert np.dot(b, b.conj()) == gemmdot(b, b, trans='c')

    # Check gemm for transa='n'
    a2 = np.arange(7 * 5 * 1 * 3).reshape(7, 5, 1, 3) * (-1. + 4.j) + 3.
    c = np.tensordot(a, a2, [1, 0])
    mmmx(1., a, 'N', a2, 'N', -1., c)
    assert not c.any()

    # Check gemm for transa='c'
    a = np.arange(4 * 5 * 1 * 3).reshape(4, 5, 1, 3) * (3. - 2.j) + 4.
    c = np.tensordot(a, a2.conj(), [[1, 2, 3], [1, 2, 3]])
    mmmx(1., a, 'N', a2, 'C', -1., c)
    assert not c.any()

    # Check axpy
    c = 5.j * a
    axpy(-5.j, a, c)
    assert not c.any()

    # Check rk
    c = np.tensordot(a, a.conj(), [[1, 2, 3], [1, 2, 3]])
    rk(1., a, -1., c)
    tri2full(c)
    assert not c.any()

    # Check gemmdot for transa='c'
    c = np.tensordot(a, a2.conj(), [-1, -1])
    gemmdot(a, a2, beta=-1., out=c, trans='c')
    assert not c.any()

    # Check gemmdot for transa='n'
    a2.shape = 3, 7, 5, 1
    c = np.tensordot(a, a2, [-1, 0])
    gemmdot(a, a2, beta=-1., out=c, trans='n')
    assert not c.any()

    # Check r2k
    a2 = 5. * a
    c = np.tensordot(a, a2.conj(), [[1, 2, 3], [1, 2, 3]])
    r2k(.5, a, a2, -1., c)
    tri2full(c)
    assert not c.any()



import pytest
from ase.lattice.hexagonal import Graphene
from ase.parallel import parprint

from gpaw import GPAW, PW, FermiDirac


@pytest.mark.legacy
def test_kpt_refine():
    system = Graphene(symbol='C',
                      latticeconstant={'a': 2.467710, 'c': 1.0},
                      size=(1, 1, 1))
    system.pbc = (1, 1, 0)
    system.center(axis=2, vacuum=4.0)

    kpt_refine = {
        'center': [1 / 3, 1 / 3, 0],
        'size': [3, 3, 1],
        'reduce_symmetry': False}
    # kpt_refine={"center":[[1./3,1./3,0.],[-1./3,-1./3,0.]], "size":[3,3,1],
    #             "reduce_symmetry":True}

    calc = GPAW(mode=PW(ecut=400),
                xc=dict(name='PBE', stencil=1),
                kpts={'size': [9, 9, 1], 'gamma': True},
                experimental={'kpt_refine': kpt_refine},
                occupations=FermiDirac(0.026))

    system.calc = calc
    energy = system.get_potential_energy()

    parprint('Energy', energy)

    assert abs(energy - -18.27) < 0.02



from __future__ import annotations

import itertools
import warnings
from math import inf
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any, Callable

import numpy as np
from gpaw.convergence_criteria import (Criterion, check_convergence,
                                       dict2criterion)
from gpaw.scf import write_iteration
from gpaw.typing import Array2D
from gpaw.yml import indent

if TYPE_CHECKING:
    from gpaw.new.calculation import DFTState


class SCFConvergenceError(Exception):
    ...


class TooFewBandsError(SCFConvergenceError):
    """Not enough bands for CBM+x convergence cfriterium."""


class SCFLoop:
    def __init__(self,
                 hamiltonian,
                 occ_calc,
                 eigensolver,
                 mixer,
                 comm,
                 convergence,
                 maxiter):
        self.hamiltonian = hamiltonian
        self.eigensolver = eigensolver
        self.mixer = mixer
        self.occ_calc = occ_calc
        self.comm = comm
        self.convergence = convergence
        self.maxiter = maxiter
        self.niter = 0
        self.update_density_and_potential = True

    def __repr__(self):
        return 'SCFLoop(...)'

    def __str__(self):
        return (f'eigensolver:\n{indent(self.eigensolver)}\n'
                f'{self.mixer}\n'
                f'occupation numbers:\n{indent(self.occ_calc)}\n')

    def iterate(self,
                state: DFTState,
                pot_calc,
                convergence=None,
                maxiter=None,
                calculate_forces=None,
                log=None):

        cc = create_convergence_criteria(convergence or self.convergence)
        maxiter = maxiter or self.maxiter

        if log:
            log('convergence criteria:')
            for criterion in cc.values():
                if criterion.description is not None:
                    log('- ' + criterion.description)
            log(f'maximum number of iterations: {self.maxiter}\n')

        self.mixer.reset()

        if self.update_density_and_potential:
            dens_error = self.mixer.mix(state.density)
        else:
            dens_error = 0.0

        for self.niter in itertools.count(start=1):
            wfs_error = self.eigensolver.iterate(state, self.hamiltonian)
            state.ibzwfs.calculate_occs(
                self.occ_calc,
                fixed_fermi_level=not self.update_density_and_potential)

            ctx = SCFContext(
                log, self.niter,
                state,
                wfs_error, dens_error,
                self.comm, calculate_forces,
                pot_calc)

            yield ctx

            converged, converged_items, entries = check_convergence(cc, ctx)
            nconverged = self.comm.sum(int(converged))
            assert nconverged in [0, self.comm.size], converged_items

            if log:
                with log.comment():
                    write_iteration(cc, converged_items, entries, ctx, log)
            if converged:
                break
            if self.niter == maxiter:
                if wfs_error < inf:
                    raise SCFConvergenceError
                raise TooFewBandsError

            if self.update_density_and_potential:
                state.density.update(state.ibzwfs,
                                     ked=pot_calc.xc.type == 'MGGA')
                dens_error = self.mixer.mix(state.density)
                state.potential, _ = pot_calc.calculate(
                    state.density, state.ibzwfs, state.potential.vHt_x)


class SCFContext:
    def __init__(self,
                 log,
                 niter: int,
                 state: DFTState,
                 wfs_error: float,
                 dens_error: float,
                 comm,
                 calculate_forces: Callable[[], Array2D],
                 pot_calc):
        self.log = log
        self.niter = niter
        self.state = state
        energy = np.array([sum(e
                               for name, e in state.potential.energies.items()
                               if name != 'stress') +
                           sum(state.ibzwfs.energies.values())])
        comm.broadcast(energy, 0)
        self.ham = SimpleNamespace(e_total_extrapolated=energy[0],
                                   get_workfunctions=self._get_workfunctions)
        self.wfs = SimpleNamespace(nvalence=state.ibzwfs.nelectrons,
                                   world=comm,
                                   eigensolver=SimpleNamespace(
                                       error=wfs_error),
                                   nspins=state.density.ndensities,
                                   collinear=state.density.collinear)
        self.dens = SimpleNamespace(
            calculate_magnetic_moments=state.density
            .calculate_magnetic_moments,
            fixed=False,
            error=dens_error)
        self.calculate_forces = calculate_forces
        self.poisson_solver = pot_calc.poisson_solver

    def _get_workfunctions(self, _):
        """
        vHt_g = self.state.vHt_x
        axes = (c, (c + 1) % 3, (c + 2) % 3)
        potential.vt_sRself.pd3.ifft(v_q, local=True).transpose(axes)
        vacuum = v_g[0].mean()
        vacuum_level =
        (fermi_level,) = self.state.ibzwfs.fermi_levels
        wf = vacuum_level - fermi_level
        delta = self.poisson_solver.correction
        return np.array([wf + 0.5 * delta, wf - 0.5 * delta])
        """


def create_convergence_criteria(criteria: dict[str, Any]
                                ) -> dict[str, Criterion]:
    for k, v in [('energy', 0.0005),        # eV / electron
                 ('density', 1.0e-4),       # electrons / electron
                 ('eigenstates', 4.0e-8)]:  # eV^2 / electron
        if k not in criteria:
            criteria[k] = v
    # Gather convergence criteria for SCF loop.
    custom = criteria.pop('custom', [])
    for name, criterion in criteria.items():
        if hasattr(criterion, 'todict'):
            # 'Copy' so no two calculators share an instance.
            criteria[name] = dict2criterion(criterion.todict())
        else:
            criteria[name] = dict2criterion({name: criterion})

    if not isinstance(custom, (list, tuple)):
        custom = [custom]
    for criterion in custom:
        if isinstance(criterion, dict):  # from .gpw file
            msg = ('Custom convergence criterion "{:s}" encountered, '
                   'which GPAW does not know how to load. This '
                   'criterion is NOT enabled; you may want to manually'
                   ' set it.'.format(criterion['name']))
            warnings.warn(msg)
            continue

        criteria[criterion.name] = criterion
        msg = ('Custom convergence criterion {:s} encountered. '
               'Please be sure that each calculator is fed a '
               'unique instance of this criterion. '
               'Note that if you save the calculator instance to '
               'a .gpw file you may not be able to re-open it. '
               .format(criterion.name))
        warnings.warn(msg)

    for criterion in criteria.values():
        criterion.reset()

    return criteria



from __future__ import annotations
from gpaw.new.lcao.eigensolver import LCAOEigensolver


class HybridXCFunctional:
    setup_name = 'PBE'

    def __init__(self, params: dict):
        name = params['name']
        print(name)

    def calculate(self, nt_sr, vxct_sr):
        # semi-local stuff here?
        ...
        return 42.0

    def calculate_paw_correction(self, setup, D_sp, dH_sp):
        ...
        return 117.0


class HybridLCAOEigensolver(LCAOEigensolver):
    def __init__(self, basis, fracpos_ac, cell_cv):
        super().__init__(basis)
        print(fracpos_ac, cell_cv)

    def iterate(self, state, hamiltonian) -> float:
        for wfs in state.ibzwfs:
            rho_MM = wfs.calculate_density_matrix()
            print(rho_MM)
        return -1.0



import numpy as np
from gpaw.response.pw_parallelization import Blocks1D
from gpaw.response.gamma_int import GammaIntegrator
from ase.utils import lazyproperty


class DielectricFunctionCalculator:
    def __init__(self, chi0, coulomb, xckernel, mode):
        self.coulomb = coulomb
        self.qpd = chi0.qpd
        self.mode = mode
        self.optical_limit = chi0.optical_limit
        self.chi0 = chi0
        self.xckernel = xckernel
        wblocks1d = Blocks1D(chi0.body.blockdist.blockcomm, len(chi0.wd))
        self.wblocks1d = wblocks1d
        # Generate fine grid in vicinity of gamma
        if chi0.optical_limit and wblocks1d.nlocal:
            self.gamma_int = GammaIntegrator(
                truncation=self.coulomb.truncation,
                kd=coulomb.kd, qpd=self.qpd,
                chi0_wvv=chi0.chi0_Wvv[wblocks1d.myslice],
                chi0_wxvG=chi0.chi0_WxvG[wblocks1d.myslice])
        else:
            self.gamma_int = None

    @lazyproperty
    def sqrtV_G(self):
        return self.coulomb.sqrtV(qpd=self.qpd, q_v=None)

    @lazyproperty
    def I_GG(self):
        return np.eye(self.qpd.ngmax)

    @lazyproperty
    def fxc_GG(self):
        if self.mode == 'GW':
            return self.I_GG
        else:
            return self.xckernel.calculate(self.qpd)

    def get_epsinv_wGG(self, only_correlation=True):
        """
        Calculates inverse dielectric matrix for all frequencies.
        """
        chi0_wGG = self.chi0.body.copy_array_with_distribution('wGG')
        epsinv_wGG = []
        for iw, chi0_GG in enumerate(chi0_wGG):
            epsinv_GG = self.get_epsinv_GG(chi0_GG, iw)
            if only_correlation:
                epsinv_GG -= self.I_GG
            epsinv_wGG.append(epsinv_GG)
        return np.asarray(epsinv_wGG)

    def get_epsinv_GG(self, chi0_GG, iw=None):
        """
        Calculates inverse dielectric matrix for single frequency
        """
        if self.optical_limit:
            return self._get_epsinvGamma_GG(chi0_GG, iw)
        dfc = _DielectricFunctionCalculator(self.sqrtV_G,
                                            chi0_GG,
                                            self.mode,
                                            self.fxc_GG)
        return dfc.get_epsinv_GG()

    def _get_epsinvGamma_GG(self, chi0_GG, iw):
        dfc = _DielectricFunctionCalculator(self.sqrtV_G,
                                            chi0_GG,
                                            self.mode,
                                            self.fxc_GG)
        return dfc.get_epsinvGamma_GG(self.gamma_int,
                                      self.qpd, iw,
                                      self.coulomb)


class _DielectricFunctionCalculator:
    def __init__(self, sqrtV_G, chi0_GG, mode, fxc_GG=None):
        self.sqrtV_G = sqrtV_G
        self.chiVV_GG = chi0_GG * sqrtV_G * sqrtV_G[:, np.newaxis]

        self.I_GG = np.eye(len(sqrtV_G))

        self.fxc_GG = fxc_GG
        self.chi0_GG = chi0_GG
        self.mode = mode

    def _chiVVfxc_GG(self):
        assert self.mode != 'GW'
        assert self.fxc_GG is not None
        return self.chiVV_GG @ self.fxc_GG

    def eps_GG_gwp(self):
        gwp_inv_GG = np.linalg.inv(self.I_GG - self._chiVVfxc_GG() +
                                   self.chiVV_GG)
        return self.I_GG - gwp_inv_GG @ self.chiVV_GG

    def eps_GG_gws(self):
        # Note how the signs are different wrt. gwp.
        # Nobody knows why.
        gws_inv_GG = np.linalg.inv(self.I_GG + self._chiVVfxc_GG() -
                                   self.chiVV_GG)
        return gws_inv_GG @ (self.I_GG - self.chiVV_GG)

    def eps_GG_plain(self):
        return self.I_GG - self.chiVV_GG

    def eps_GG_w_fxc(self):
        return self.I_GG - self._chiVVfxc_GG()

    def get_eps_GG(self):
        mode = self.mode
        if mode == 'GWP':
            return self.eps_GG_gwp()
        elif mode == 'GWS':
            return self.eps_GG_gws()
        elif mode == 'GW':
            return self.eps_GG_plain()
        elif mode == 'GWG':
            return self.eps_GG_w_fxc()
        raise ValueError(f'Unknown mode: {mode}')

    def get_epsinv_GG(self):
        eps_GG = self.get_eps_GG()
        return np.linalg.inv(eps_GG)

    def get_epsinvGamma_GG(self, gamma_int, qpd, iw, coulomb):
        # Get average epsinv over small region around Gamma
        epsinv_GG = np.zeros(self.chi0_GG.shape, complex)
        for iqf in range(len(gamma_int.qf_qv)):
            # Note! set_appendages changes (0,0), (0,:), (:,0)
            # elements of chi0_GG
            gamma_int.set_appendages(self.chi0_GG, iw, iqf)
            sqrtV_G = coulomb.sqrtV(
                qpd=qpd, q_v=gamma_int.qf_qv[iqf])
            dfc = _DielectricFunctionCalculator(
                sqrtV_G, self.chi0_GG, mode=self.mode, fxc_GG=self.fxc_GG)
            epsinv_GG += dfc.get_epsinv_GG() * gamma_int.weight_q
        return epsinv_GG



from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal


def test_generic_IP_oxygen():
    a = 6.0
    O = Atoms('O',
              [(a / 2, a / 2 + 0.5, a / 2)],
              magmoms=[2],
              pbc=False,
              cell=(a, a + 1, a))
    O.calc = GPAW(mode='fd', gpts=(32, 36, 32), nbands=4)
    e0 = O.get_potential_energy()

    O.calc = GPAW(mode='fd', gpts=(32, 36, 32), nbands=4, charge=1)

    e1 = O.get_potential_energy()

    print(e1 - e0)
    assert abs(e1 - e0 - 13.989) < 0.04

    energy_tolerance = 0.004
    equal(e0, -1.88477, energy_tolerance)
    equal(e1, 12.11080, energy_tolerance)

    # The first ionization energy for LDA oxygen is from this paper:
    # In-Ho Lee, Richard M. Martin, Phys. Rev. B 56 7197 (1997)



from math import pi

from ase import Atoms
from ase.parallel import parprint

from gpaw import GPAW
from gpaw.test import equal
from gpaw.analyse.expandyl import AngularIntegral, ExpandYl


def test_radial_ylexpand(in_tmp_dir):
    fname = 'H2.gpw'
    donot = ''
    donot = 'donot'
    try:
        calc = GPAW(fname + donot, txt=None)
        H2 = calc.get_atoms()
        calc.converge_wave_functions()
    except FileNotFoundError:
        R = 0.7  # approx. experimental bond length
        a = 2.
        c = 3.
        H2 = Atoms('HH',
                   [(a / 2, a / 2, (c - R) / 2),
                    (a / 2, a / 2, (c + R) / 2)],
                   cell=(a, a, c),
                   pbc=True)
        calc = GPAW(mode='fd', gpts=(12, 12, 16), nbands=2, kpts=(1, 1, 2),
                    convergence={'eigenstates': 1.e-6},
                    txt=None,
                    )
        H2.calc = calc
        H2.get_potential_energy()
        if not donot:
            calc.write(fname)

    # Check that a / h = 10 is rounded up to 12 as always:
    assert (calc.wfs.gd.N_c == (12, 12, 16)).all()

    # AngularIntegral

    gd = calc.density.gd
    ai = AngularIntegral(H2.positions.mean(0), calc.wfs.gd, Rmax=1.5)
    unity_g = gd.zeros() + 1.
    average_R = ai.average(unity_g)
    integral_R = ai.integrate(unity_g)
    for V, average, integral, R, Rm in zip(ai.V_R, average_R, integral_R,
                                           ai.radii(), ai.radii('mean')):
        if V > 0:
            equal(average, 1, 1.e-9)
            equal(integral / (4 * pi * Rm**2), 1, 0.61)
            equal(Rm / R, 1, 0.61)

    # ExpandYl

    yl = ExpandYl(H2.positions.mean(0), calc.wfs.gd, Rmax=1.5)

    def max_index(l):
        mi = 0
        limax = l[0]
        for i, li in enumerate(l):
            if limax < li:
                limax = li
                mi = i
        return mi

    # check numbers
    for n in [0, 1]:
        # gl, w = yl.expand(calc.get_pseudo_wave_function(band=n))
        gl, w = yl.expand(calc.wfs.kpt_u[0].psit_nG[n])
        parprint('max_index(gl), n=', max_index(gl), n)
        assert max_index(gl) == n

    # io
    fname = 'expandyl.dat'
    yl.to_file(calc, fname)



import numpy as np
import pytest
from ase.calculators.test import numeric_force
from ase.parallel import parprint

from gpaw import GPAW


@pytest.mark.mgga
def test_pw_si_stress_mgga(gpw_files, gpaw_new):
    si = GPAW(gpw_files['si_pw_distorted']).get_atoms()

    # Trigger nasty bug (fixed in !486):
    if not gpaw_new:
        si.calc.wfs.pt.blocksize = si.calc.wfs.pd.maxmyng - 1

    s_analytical = si.get_stress()
    # Calculated numerical stress once, store here to speed up test
    s_numerical = np.array([-0.01140242, -0.04084746, -0.0401058,
                            -0.02119629, 0.13584242, 0.00911572])
    if 0:
        s_numerical = si.calc.calculate_numerical_stress(si, 1e-5)

    s_err = s_numerical - s_analytical

    parprint('Analytical stress:\n', s_analytical)
    parprint('Numerical stress:\n', s_numerical)
    parprint('Error in stress:\n', s_err)
    assert np.all(abs(s_err) < 1e-5)

    # Check y-component of second atom:
    f = si.get_forces()[1, 1]
    fref = -2.066952082010687
    if 0:
        fref = numeric_force(si, 1, 1)
    print(f, fref, f - fref)
    assert abs(f - fref) < 0.0005



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""
See Kresse, Phys. Rev. B 54, 11169 (1996)
"""

import numpy as np
from numpy.fft import fftn, ifftn

import gpaw.mpi as mpi
from gpaw.utilities.blas import axpy
from gpaw.fd_operators import FDOperator
from gpaw.utilities.tools import construct_reciprocal

"""About mixing-related classes.

(FFT/Broyden)BaseMixer: These classes know how to mix one density
array and store history etc.  But they do not take care of complexity
like spin.

(SpinSum/etc.)MixerDriver: These combine one or more BaseMixers to
implement a full algorithm.  Think of them as stateless (immutable).
The user can give an object of these types as input, but they will generally
be constructed by a utility function so the interface is nice.

The density object always wraps the (X)MixerDriver with a
MixerWrapper.  The wrapper contains the common code for all mixers so
we don't have to implement it multiple times (estimate memory, etc.).

In the end, what the user provides is probably a dictionary anyway, and the
relevant objects are instantiated automatically."""


class BaseMixer:
    name = 'pulay'

    """Pulay density mixer."""
    def __init__(self, beta, nmaxold, weight):
        """Construct density-mixer object.

        Parameters:

        beta: float
            Mixing parameter between zero and one (one is most
            aggressive).
        nmaxold: int
            Maximum number of old densities.
        weight: float
            Weight parameter for special metric (for long wave-length
            changes).

        """

        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight
        self.world = None

    def initialize_metric(self, gd):
        self.gd = gd

        if self.weight == 1:
            self.metric = None
        else:
            a = 0.125 * (self.weight + 7)
            b = 0.0625 * (self.weight - 1)
            c = 0.03125 * (self.weight - 1)
            d = 0.015625 * (self.weight - 1)
            self.metric = FDOperator([a,
                                      b, b, b, b, b, b,
                                      c, c, c, c, c, c, c, c, c, c, c, c,
                                      d, d, d, d, d, d, d, d],
                                     [(0, 0, 0),  # a
                                      (-1, 0, 0), (1, 0, 0),  # b
                                      (0, -1, 0), (0, 1, 0),
                                      (0, 0, -1), (0, 0, 1),
                                      (1, 1, 0), (1, 0, 1), (0, 1, 1),  # c
                                      (1, -1, 0), (1, 0, -1), (0, 1, -1),
                                      (-1, 1, 0), (-1, 0, 1), (0, -1, 1),
                                      (-1, -1, 0), (-1, 0, -1), (0, -1, -1),
                                      (1, 1, 1), (1, 1, -1), (1, -1, 1),  # d
                                      (-1, 1, 1), (1, -1, -1), (-1, -1, 1),
                                      (-1, 1, -1), (-1, -1, -1)],
                                     gd, float).apply
            self.mR_sG = gd.empty(4)

    def reset(self):
        """Reset Density-history.

        Called at initialization and after each move of the atoms.

        my_nuclei:   All nuclei in local domain.
        """

        # History for Pulay mixing of densities:
        self.nt_isG = []  # Pseudo-electron densities
        self.R_isG = []  # Residuals
        self.A_ii = np.zeros((0, 0))

        self.D_iasp = []
        self.dD_iasp = []

    def calculate_charge_sloshing(self, R_sG) -> float:
        return self.gd.integrate(np.fabs(R_sG)).sum()

    def mix_density(self, nt_sG, D_asp, g_ss=None):
        nt_isG = self.nt_isG
        R_isG = self.R_isG
        D_iasp = self.D_iasp
        dD_iasp = self.dD_iasp
        spin = len(nt_sG)
        iold = len(self.nt_isG)
        dNt = np.inf
        if iold > 0:
            if iold > self.nmaxold:
                # Throw away too old stuff:
                del nt_isG[0]
                del R_isG[0]
                del D_iasp[0]
                del dD_iasp[0]
                # for D_p, D_ip, dD_ip in self.D_a:
                #     del D_ip[0]
                #     del dD_ip[0]
                iold = self.nmaxold

            # Calculate new residual (difference between input and
            # output density):
            R_sG = nt_sG - nt_isG[-1]
            dNt = self.calculate_charge_sloshing(R_sG)
            R_isG.append(R_sG)
            dD_iasp.append([])
            for D_sp, D_isp in zip(D_asp, D_iasp[-1]):
                dD_iasp[-1].append(D_sp - D_isp)

            if self.metric is None:
                mR_sG = R_sG
            else:
                mR_sG = self.mR_sG[:spin]
                for s in range(spin):
                    self.metric(R_sG[s], mR_sG[s])

            if g_ss is not None:
                mR_sG = np.tensordot(g_ss, mR_sG, axes=(1, 0))

            # Update matrix:
            A_ii = np.zeros((iold, iold))
            i2 = iold - 1

            for i1, R_1sG in enumerate(R_isG):
                a = self.gd.comm.sum(self.dotprod(R_1sG, mR_sG, dD_iasp[i1],
                                                  dD_iasp[-1]))
                A_ii[i1, i2] = a
                A_ii[i2, i1] = a
            A_ii[:i2, :i2] = self.A_ii[-i2:, -i2:]
            self.A_ii = A_ii

            try:
                B_ii = np.linalg.inv(A_ii)
                alpha_i = B_ii.sum(1)
                alpha_i /= alpha_i.sum()
            except (ZeroDivisionError, np.linalg.LinAlgError):
                alpha_i = np.zeros(iold)
                alpha_i[-1] = 1.0

            if self.world:
                self.world.broadcast(alpha_i, 0)

            # Calculate new input density:
            nt_sG[:] = 0.0
            # for D_p, D_ip, dD_ip in self.D_a:
            for D in D_asp:
                D[:] = 0.0
            beta = self.beta
            for i, alpha in enumerate(alpha_i):
                axpy(alpha, nt_isG[i], nt_sG)
                axpy(alpha * beta, R_isG[i], nt_sG)

                for D_sp, D_isp, dD_isp in zip(D_asp, D_iasp[i],
                                               dD_iasp[i]):
                    axpy(alpha, D_isp, D_sp)
                    axpy(alpha * beta, dD_isp, D_sp)

        # Store new input density (and new atomic density matrices):
        nt_isG.append(nt_sG.copy())
        D_iasp.append([])
        for D_sp in D_asp:
            D_iasp[-1].append(D_sp.copy())
        return dNt

    # may presently be overridden by passing argument in constructor
    def dotprod(self, R1_G, R2_G, dD1_ap, dD2_ap):
        return np.vdot(R1_G, R2_G).real

    def estimate_memory(self, mem, gd):
        gridbytes = gd.bytecount()
        mem.subnode('nt_iG, R_iG', 2 * self.nmaxold * gridbytes)

    def __repr__(self):
        classname = self.__class__.__name__
        template = '%s(beta=%f, nmaxold=%d, weight=%f)'
        string = template % (classname, self.beta, self.nmaxold, self.weight)
        return string


class ExperimentalDotProd:
    def __init__(self, calc):
        self.calc = calc

    def __call__(self, R1_G, R2_G, dD1_ap, dD2_ap):
        prod = np.vdot(R1_G, R2_G).real
        setups = self.calc.wfs.setups
        # okay, this is a bit nasty because it depends on dD1_ap
        # and its friend having come from D_asp.values() and the dictionaries
        # not having been modified.  This is probably true... for now.
        avalues = self.calc.density.D_asp.keys()
        for a, dD1_p, dD2_p in zip(avalues, dD1_ap, dD2_ap):
            I4_pp = setups[a].four_phi_integrals()
            dD4_pp = np.outer(dD1_p, dD2_p)  # not sure if corresponds quite
            prod += (I4_pp * dD4_pp).sum()
        return prod


class ReciprocalMetric:
    def __init__(self, weight, k2_Q):
        self.k2_Q = k2_Q
        k2_min = np.min(self.k2_Q)
        self.q1 = (weight - 1) * k2_min

    def __call__(self, R_Q, mR_Q):
        mR_Q[:] = R_Q * (1.0 + self.q1 / self.k2_Q)


class FFTBaseMixer(BaseMixer):
    name = 'fft'

    """Mix the density in Fourier space"""
    def __init__(self, beta, nmaxold, weight):
        BaseMixer.__init__(self, beta, nmaxold, weight)
        self.gd1 = None

    def initialize_metric(self, gd):
        self.gd = gd

        if gd.comm.rank == 0:
            self.gd1 = gd.new_descriptor(comm=mpi.serial_comm)
            k2_Q, _ = construct_reciprocal(self.gd1)
            self.metric = ReciprocalMetric(self.weight, k2_Q)
            self.mR_sG = self.gd1.empty(2, dtype=complex)
        else:
            self.metric = lambda R_Q, mR_Q: None
            self.mR_sG = np.empty((2, 0, 0, 0), dtype=complex)

    def calculate_charge_sloshing(self, R_sQ):
        if self.gd.comm.rank == 0:
            assert R_sQ.ndim == 4  # and len(R_sQ) == 1
            cs = sum(self.gd1.integrate(np.fabs(ifftn(R_Q).real))
                     for R_Q in R_sQ)
        else:
            cs = 0.0
        return self.gd.comm.sum(cs)

    def mix_density(self, nt_sR, D_asp, g_ss=None):
        # Transform real-space density to Fourier space
        nt1_sR = [self.gd.collect(nt_R) for nt_R in nt_sR]
        if self.gd.comm.rank == 0:
            nt1_sG = np.ascontiguousarray([fftn(nt_R) for nt_R in nt1_sR])
        else:
            nt1_sG = np.empty((len(nt_sR), 0, 0, 0), dtype=complex)

        dNt = BaseMixer.mix_density(self, nt1_sG, D_asp)

        # Return density in real space
        for nt_G, nt_R in zip(nt1_sG, nt_sR):
            if self.gd.comm.rank == 0:
                nt1_R = ifftn(nt_G).real
            else:
                nt1_R = None
            self.gd.distribute(nt1_R, nt_R)

        return dNt


class BroydenBaseMixer:
    name = 'broyden'

    def __init__(self, beta, nmaxold, weight):
        self.verbose = False
        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = 1.0  # XXX discards argument

    def initialize_metric(self, gd):
        self.gd = gd

    def reset(self):
        self.step = 0
        # self.d_nt_G = []
        # self.d_D_ap = []

        self.R_iG = []
        self.dD_iap = []

        self.nt_iG = []
        self.D_iap = []
        self.c_G = []
        self.v_G = []
        self.u_G = []
        self.u_D = []

    def mix_density(self, nt_sG, D_asp):
        nt_G = nt_sG[0]
        D_ap = [D_sp[0] for D_sp in D_asp]
        dNt = np.inf
        if self.step > 2:
            del self.R_iG[0]
            for d_Dp in self.dD_iap:
                del d_Dp[0]
        if self.step > 0:
            self.R_iG.append(nt_G - self.nt_iG[-1])
            for d_Dp, D_p, D_ip in zip(self.dD_iap, D_ap, self.D_iap):
                d_Dp.append(D_p - D_ip[-1])
            fmin_G = self.gd.integrate(self.R_iG[-1] * self.R_iG[-1])
            dNt = self.gd.integrate(np.fabs(self.R_iG[-1]))
            if self.verbose:
                print('Mixer: broyden: fmin_G = %f fmin_D = %f' % fmin_G)
        if self.step == 0:
            self.eta_G = np.empty(nt_G.shape)
            self.eta_D = []
            for D_p in D_ap:
                self.eta_D.append(0)
                self.u_D.append([])
                self.D_iap.append([])
                self.dD_iap.append([])
        else:
            if self.step >= 2:
                del self.c_G[:]
                if len(self.v_G) >= self.nmaxold:
                    del self.u_G[0]
                    del self.v_G[0]
                    for u_D in self.u_D:
                        del u_D[0]
                temp_nt_G = self.R_iG[1] - self.R_iG[0]
                self.v_G.append(temp_nt_G / self.gd.integrate(temp_nt_G *
                                                              temp_nt_G))
                if len(self.v_G) < self.nmaxold:
                    nstep = self.step - 1
                else:
                    nstep = self.nmaxold
                for i in range(nstep):
                    self.c_G.append(self.gd.integrate(self.v_G[i] *
                                                      self.R_iG[1]))
                self.u_G.append(self.beta * temp_nt_G + self.nt_iG[1] -
                                self.nt_iG[0])
                for d_Dp, u_D, D_ip in zip(self.dD_iap, self.u_D, self.D_iap):
                    temp_D_ap = d_Dp[1] - d_Dp[0]
                    u_D.append(self.beta * temp_D_ap + D_ip[1] - D_ip[0])
                usize = len(self.u_G)
                for i in range(usize - 1):
                    a_G = self.gd.integrate(self.v_G[i] * temp_nt_G)
                    axpy(-a_G, self.u_G[i], self.u_G[usize - 1])
                    for u_D in self.u_D:
                        axpy(-a_G, u_D[i], u_D[usize - 1])
            self.eta_G = self.beta * self.R_iG[-1]
            for i, d_Dp in enumerate(self.dD_iap):
                self.eta_D[i] = self.beta * d_Dp[-1]
            usize = len(self.u_G)
            for i in range(usize):
                axpy(-self.c_G[i], self.u_G[i], self.eta_G)
                for eta_D, u_D in zip(self.eta_D, self.u_D):
                    axpy(-self.c_G[i], u_D[i], eta_D)
            axpy(-1.0, self.R_iG[-1], nt_G)
            axpy(1.0, self.eta_G, nt_G)
            for D_p, d_Dp, eta_D in zip(D_ap, self.dD_iap, self.eta_D):
                axpy(-1.0, d_Dp[-1], D_p)
                axpy(1.0, eta_D, D_p)
            if self.step >= 2:
                del self.nt_iG[0]
                for D_ip in self.D_iap:
                    del D_ip[0]
        self.nt_iG.append(np.copy(nt_G))
        for D_ip, D_p in zip(self.D_iap, D_ap):
            D_ip.append(np.copy(D_p))
        self.step += 1
        return dNt


class DummyMixer:
    """Dummy mixer for TDDFT, i.e., it does not mix."""
    name = 'dummy'
    beta = 1.0
    nmaxold = 1
    weight = 1

    def __init__(self, *args, **kwargs):
        return

    def mix(self, basemixers, nt_sG, D_asp):
        return 0.0

    def get_basemixers(self, nspins):
        return []

    def todict(self):
        return {'name': 'dummy'}


class NotMixingMixer:
    name = 'no-mixing'

    def __init__(self, beta, nmaxold, weight):
        """Construct density-mixer object.
        Parameters: they are ignored for this mixer
        """

        # whatever parameters you give it doesn't do anything with them
        self.beta = 0
        self.nmaxold = 0
        self.weight = 0

    def initialize_metric(self, gd):
        self.gd = gd
        self.metric = None

    def reset(self):
        """Reset Density-history.

        Called at initialization and after each move of the atoms.

        my_nuclei:   All nuclei in local domain.
        """

        # Previous density:
        self.nt_isG = []  # Pseudo-electron densities

    def calculate_charge_sloshing(self, R_sG):
        return self.gd.integrate(np.fabs(R_sG)).sum()

    def mix_density(self, nt_sG, D_asp):
        iold = len(self.nt_isG)

        dNt = np.inf
        if iold > 0:
            # Calculate new residual (difference between input and
            # output density):
            dNt = self.calculate_charge_sloshing(nt_sG - self.nt_isG[-1])
        # Store new input density:
        self.nt_isG = [nt_sG.copy()]

        return dNt

    # may presently be overridden by passing argument in constructor
    def dotprod(self, R1_G, R2_G, dD1_ap, dD2_ap):
        pass

    def estimate_memory(self, mem, gd):
        gridbytes = gd.bytecount()
        mem.subnode('nt_iG, R_iG', 2 * self.nmaxold * gridbytes)

    def __repr__(self):
        string = 'no mixing of density'
        return string


class SeparateSpinMixerDriver:
    name = 'separate'

    def __init__(self, basemixerclass, beta, nmaxold, weight):
        self.basemixerclass = basemixerclass

        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight

    def get_basemixers(self, nspins):
        return [self.basemixerclass(self.beta, self.nmaxold, self.weight)
                for _ in range(nspins)]

    def mix(self, basemixers, nt_sG, D_asp):
        """Mix pseudo electron densities."""
        D_asp = D_asp.values()
        dNt = 0.0
        for s, (nt_G, basemixer) in enumerate(zip(nt_sG, basemixers)):
            D_a1p = [D_sp[s:s + 1] for D_sp in D_asp]
            nt_1G = nt_G[np.newaxis]
            dNt += basemixer.mix_density(nt_1G, D_a1p)
        return dNt


class SpinSumMixerDriver:
    name = 'sum'
    mix_atomic_density_matrices = False

    def __init__(self, basemixerclass, beta, nmaxold, weight):
        self.basemixerclass = basemixerclass

        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight

    def get_basemixers(self, nspins):
        if nspins == 1:
            raise ValueError('Spin sum mixer expects 2 or 4 components')
        return [self.basemixerclass(self.beta, self.nmaxold, self.weight)]

    def mix(self, basemixers, nt_sG, D_asp):
        assert len(basemixers) == 1
        basemixer = basemixers[0]
        D_asp = D_asp.values()

        collinear = len(nt_sG) == 2

        # Mix density
        if collinear:
            nt_1G = nt_sG.sum(0)[np.newaxis]
        else:
            nt_1G = nt_sG[:1]

        if self.mix_atomic_density_matrices:
            if collinear:
                D_a1p = [D_sp[:1] + D_sp[1:] for D_sp in D_asp]
            else:
                D_a1p = [D_sp[:1] for D_sp in D_asp]
            dNt = basemixer.mix_density(nt_1G, D_a1p)
            if collinear:
                dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]
                for D_sp, D_1p, dD_p in zip(D_asp, D_a1p, dD_ap):
                    D_sp[0] = 0.5 * (D_1p[0] + dD_p)
                    D_sp[1] = 0.5 * (D_1p[0] - dD_p)
        else:
            dNt = basemixer.mix_density(nt_1G, D_asp)

        if collinear:
            dnt_G = nt_sG[0] - nt_sG[1]
            # Only new magnetization for spin density
            # dD_ap = [D_sp[0] - D_sp[1] for D_sp in D_asp]

            # Construct new spin up/down densities
            nt_sG[0] = 0.5 * (nt_1G[0] + dnt_G)
            nt_sG[1] = 0.5 * (nt_1G[0] - dnt_G)

        return dNt


class SpinSumMixerDriver2(SpinSumMixerDriver):
    name = 'sum2'
    mix_atomic_density_matrices = True


class SpinDifferenceMixerDriver:
    name = 'difference'

    def __init__(self, basemixerclass, beta, nmaxold, weight,
                 beta_m=0.7, nmaxold_m=2, weight_m=10.0):
        self.basemixerclass = basemixerclass
        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight
        self.beta_m = beta_m
        self.nmaxold_m = nmaxold_m
        self.weight_m = weight_m

    def get_basemixers(self, nspins):
        if nspins == 1:
            raise ValueError('Spin difference mixer expects 2 or 4 components')
        basemixer = self.basemixerclass(self.beta, self.nmaxold, self.weight)
        if nspins == 2:
            basemixer_m = self.basemixerclass(self.beta_m, self.nmaxold_m,
                                              self.weight_m)
            return basemixer, basemixer_m
        else:
            basemixer_x = self.basemixerclass(self.beta_m, self.nmaxold_m,
                                              self.weight_m)
            basemixer_y = self.basemixerclass(self.beta_m, self.nmaxold_m,
                                              self.weight_m)
            basemixer_z = self.basemixerclass(self.beta_m, self.nmaxold_m,
                                              self.weight_m)
            return basemixer, basemixer_x, basemixer_y, basemixer_z

    def mix(self, basemixers, nt_sG, D_asp):
        D_asp = D_asp.values()

        if len(nt_sG) == 2:
            basemixer, basemixer_m = basemixers
        else:
            assert len(nt_sG) == 4
            basemixer, basemixer_x, basemixer_y, basemixer_z = basemixers

        if len(nt_sG) == 2:
            # Mix density
            nt_1G = nt_sG.sum(0)[np.newaxis]
            D_a1p = [D_sp[:1] + D_sp[1:] for D_sp in D_asp]
            dNt = basemixer.mix_density(nt_1G, D_a1p)

            # Mix magnetization
            dnt_1G = nt_sG[:1] - nt_sG[1:]
            dD_a1p = [D_sp[:1] - D_sp[1:] for D_sp in D_asp]
            basemixer_m.mix_density(dnt_1G, dD_a1p)
            # (The latter is not counted in dNt)

            # Construct new spin up/down densities
            nt_sG[:1] = 0.5 * (nt_1G + dnt_1G)
            nt_sG[1:] = 0.5 * (nt_1G - dnt_1G)
            for D_sp, D_1p, dD_1p in zip(D_asp, D_a1p, dD_a1p):
                D_sp[:1] = 0.5 * (D_1p + dD_1p)
                D_sp[1:] = 0.5 * (D_1p - dD_1p)
        else:
            # Mix density
            nt_1G = nt_sG[:1]
            D_a1p = [D_sp[:1] for D_sp in D_asp]
            dNt = basemixer.mix_density(nt_1G, D_a1p)

            # Mix magnetization
            Dx_a1p = [D_sp[1:2] for D_sp in D_asp]
            Dy_a1p = [D_sp[2:3] for D_sp in D_asp]
            Dz_a1p = [D_sp[3:4] for D_sp in D_asp]

            basemixer_x.mix_density(nt_sG[1:2], Dx_a1p)
            basemixer_y.mix_density(nt_sG[2:3], Dy_a1p)
            basemixer_z.mix_density(nt_sG[3:4], Dz_a1p)
        return dNt


class FullSpinMixerDriver:
    name = 'fullspin'

    def __init__(self, basemixerclass, beta, nmaxold, weight, g=None):
        self.basemixerclass = basemixerclass
        self.beta = beta
        self.nmaxold = nmaxold
        self.weight = weight
        self.g_ss = g

    def get_basemixers(self, nspins):
        if nspins == 1:
            raise ValueError('Full-spin mixer expects 2 or 4 spin channels')

        basemixer = self.basemixerclass(self.beta, self.nmaxold, self.weight)
        return [basemixer]

    def mix(self, basemixers, nt_sG, D_asp):
        D_asp = D_asp.values()
        basemixer = basemixers[0]
        if self.g_ss is None:
            self.g_ss = np.identity(len(nt_sG))

        dNt = basemixer.mix_density(nt_sG, D_asp, self.g_ss)

        return dNt


# Dictionaries to get mixers by name:
_backends = {}
_methods = {}
for cls in [FFTBaseMixer, BroydenBaseMixer, BaseMixer, NotMixingMixer]:
    _backends[cls.name] = cls  # type:ignore
for dcls in [SeparateSpinMixerDriver, SpinSumMixerDriver,
             FullSpinMixerDriver, SpinSumMixerDriver2,
             SpinDifferenceMixerDriver, DummyMixer]:
    _methods[dcls.name] = dcls  # type:ignore


# This function is used by Density to decide mixer parameters
# that the user did not explicitly provide, i.e., it fills out
# everything that is missing and returns a mixer "driver".
def get_mixer_from_keywords(pbc, nspins, **mixerkwargs):
    if mixerkwargs.get('name') == 'dummy':
        return DummyMixer()

    if mixerkwargs.get('backend') == 'no-mixing':
        mixerkwargs['beta'] = 0
        mixerkwargs['nmaxold'] = 0
        mixerkwargs['weight'] = 0

    if nspins == 1:
        mixerkwargs['method'] = SeparateSpinMixerDriver

    # The plan is to first establish a kwargs dictionary with all the
    # defaults, then we update it with values from the user.
    kwargs = {'backend': BaseMixer}

    if np.any(pbc):  # Works on array or boolean
        kwargs.update(beta=0.05, history=5, weight=50.0)
    else:
        kwargs.update(beta=0.25, history=3, weight=1.0)

    if nspins == 1:
        kwargs['method'] = SeparateSpinMixerDriver
    else:
        kwargs['method'] = SpinDifferenceMixerDriver

    # Clean up mixerkwargs (compatibility)
    if 'nmaxold' in mixerkwargs:
        assert 'history' not in mixerkwargs
        mixerkwargs['history'] = mixerkwargs.pop('nmaxold')

    # Now the user override:
    for key in kwargs:
        # Clean any 'None' values out as if they had never been passed:
        val = mixerkwargs.pop(key, None)
        if val is not None:
            kwargs[key] = val

    # Resolve keyword strings (like 'fft') into classes (like FFTBaseMixer):
    driver = _methods.get(kwargs['method'], kwargs['method'])
    baseclass = _backends.get(kwargs['backend'], kwargs['backend'])

    # We forward any remaining mixer kwargs to the actual mixer object.
    # Any user defined variables that do not really exist will cause an error.
    mixer = driver(baseclass, beta=kwargs['beta'],
                   nmaxold=kwargs['history'], weight=kwargs['weight'],
                   **mixerkwargs)
    return mixer


# This is the only object which will be used by Density, sod the others
class MixerWrapper:
    def __init__(self, driver, nspins, gd, world=None):
        self.driver = driver

        self.beta = driver.beta
        self.nmaxold = driver.nmaxold
        self.weight = driver.weight
        assert self.weight is not None, driver

        self.basemixers = self.driver.get_basemixers(nspins)
        for basemixer in self.basemixers:
            basemixer.initialize_metric(gd)
            basemixer.world = world

    def mix(self, nt_sR, D_asp=None):
        if D_asp is not None:
            return self.driver.mix(self.basemixers, nt_sR, D_asp)

        # new interface:
        density = nt_sR
        nspins = density.nt_sR.dims[0]
        nt_sR = density.nt_sR.to_xp(np)
        D_asii = density.D_asii.to_xp(np)
        D_asp = {a: D_sii.copy().reshape((nspins, -1))
                 for a, D_sii in D_asii.items()}
        error = self.driver.mix(self.basemixers,
                                nt_sR.data,
                                D_asp)
        for a, D_sii in D_asii.items():
            ni = D_sii.shape[1]
            D_sii[:] = D_asp[a].reshape((nspins, ni, ni))
        xp = density.nt_sR.xp
        if xp is not np:
            density.nt_sR.data[:] = xp.asarray(nt_sR.data)
            density.D_asii.data[:] = xp.asarray(D_asii.data)
        return error

    def estimate_memory(self, mem, gd):
        for i, basemixer in enumerate(self.basemixers):
            basemixer.estimate_memory(mem.subnode('Mixer %d' % i), gd)

    def reset(self):
        for basemixer in self.basemixers:
            basemixer.reset()

    def __str__(self):
        lines = ['Density mixing:',
                 'Method: ' + self.driver.name,
                 'Backend: ' + self.driver.basemixerclass.name,
                 'Linear mixing parameter: %g' % self.beta,
                 f'old densities: {self.nmaxold}',
                 'Damping of long wavelength oscillations: %g' % self.weight]
        if self.weight == 1:
            lines[-1] += '  # (no daming)'
        return '\n  '.join(lines)


# Helper function to define old-style interfaces to mixers.
# Defines and returns a function which looks like a mixer class
def _definemixerfunc(method, backend):
    def getmixer(beta=None, nmaxold=None, weight=None, **kwargs):
        d = dict(method=method, backend=backend,
                 beta=beta, nmaxold=nmaxold, weight=weight)
        d.update(kwargs)
        return d
    return getmixer


Mixer = _definemixerfunc('separate', 'pulay')
MixerSum = _definemixerfunc('sum', 'pulay')
MixerSum2 = _definemixerfunc('sum2', 'pulay')
MixerDif = _definemixerfunc('difference', 'pulay')
MixerFull = _definemixerfunc('fullspin', 'pulay')
FFTMixer = _definemixerfunc('separate', 'fft')
FFTMixerSum = _definemixerfunc('sum', 'fft')
FFTMixerSum2 = _definemixerfunc('sum2', 'fft')
FFTMixerDif = _definemixerfunc('difference', 'fft')
FFTMixerFull = _definemixerfunc('fullspin', 'fft')
BroydenMixer = _definemixerfunc('separate', 'broyden')
BroydenMixerSum = _definemixerfunc('sum', 'broyden')
BroydenMixerSum2 = _definemixerfunc('sum2', 'broyden')
BroydenMixerDif = _definemixerfunc('difference', 'broyden')



# Copyright (C) 2010  CAMd
# Copyright (C) 2010  Argonne National Laboratory
# Please see the accompanying LICENSE file for further information.

"""Module for high-level BLACS interface.

Usage
=====

A BLACS grid is a logical grid of processors.  To use BLACS, first
create a BLACS grid.  If comm contains 8 or more ranks, this example
will work::

  from gpaw.mpi import world
  from gpaw.blacs import BlacsGrid
  grid = BlacsGrid(world, 4, 2)

Use the processor grid to create various descriptors for distributed
arrays::

  block_desc = grid.new_descriptor(500, 500, 64, 64)
  local_desc = grid.new_descriptor(500, 500, 500, 500)

The first descriptor describes 500 by 500 arrays distributed amongst
the 8 CPUs of the BLACS grid in blocks of 64 by 64 elements (which is
a sensible block size).  That means each CPU has many blocks located
all over the array::

  print(world.rank, block_desc.shape, block_desc.gshape)

Here block_desc.shape is the local array shape while gshape is the
global shape.  The local array shape varies a bit on each CPU as the
block distribution may be slightly uneven.

The second descriptor, local_desc, has a block size equal to the
global size of the array, and will therefore only have one block.
This block will then reside on the first CPU -- local_desc therefore
represents non-distributed arrays.  Let us instantiate some arrays::

  H_MM = local_desc.empty()

  if world.rank == 0:
      assert H_MM.shape == (500, 500)
      H_MM[:, :] = calculate_hamiltonian_or_something()
  else:
      assert H_MM.shape[0] == 0 or H_MM.shape[1] == 0

  H_mm = block_desc.empty()
  print(H_mm.shape)  # many elements on all CPUs

We can then redistribute the local H_MM into H_mm::

  from gpaw.blacs import Redistributor
  redistributor = Redistributor(world, local_desc, block_desc)
  redistributor.redistribute(H_MM, H_mm)

Now we can run parallel linear algebra on H_mm.  This will diagonalize
H_mm, place the eigenvectors in C_mm and the eigenvalues globally in
eps_M::

  eps_M = np.empty(500)
  C_mm = block_desc.empty()
  block_desc.diagonalize_ex(H_mm, C_mm, eps_M)

We can redistribute C_mm back to the master process if we want::

  C_MM = local_desc.empty()
  redistributor2 = Redistributor(world, block_desc, local_desc)
  redistributor2.redistribute(C_mm, C_MM)

If somebody wants to do all this more easily, they will probably write
a function for that.

List of interesting classes
===========================

 * BlacsGrid
 * BlacsDescriptor
 * Redistributor

The other classes in this module are coded specifically for GPAW and
are inconvenient to use otherwise.

The module gpaw.utilities.blacs contains several functions like gemm,
gemv and r2k.  These functions may or may not have appropriate
docstings, and may use Fortran-like variable naming.  Also, either
this module or gpaw.utilities.blacs will be renamed at some point.

"""

import numpy as np

import gpaw
from gpaw.mpi import SerialCommunicator
from gpaw.matrix_descriptor import MatrixDescriptor
from gpaw.utilities.scalapack import scalapack_inverse_cholesky, \
    scalapack_diagonalize_ex, scalapack_general_diagonalize_ex, \
    scalapack_diagonalize_dc, scalapack_general_diagonalize_dc, \
    scalapack_diagonalize_mr3, scalapack_general_diagonalize_mr3
import _gpaw


INACTIVE = -1
BLOCK_CYCLIC_2D = 1


class BlacsGrid:
    """Class representing a 2D grid of processors sharing a Blacs context.

    A BLACS grid defines a logical M by N ordering of a collection of
    CPUs.  A BLACS grid can be used to create BLACS descriptors.  On
    an npcol by nprow BLACS grid, a matrix is distributed amongst M by
    N CPUs along columns and rows, respectively, while the matrix
    shape and blocking properties are determined by the descriptors.

    Use the method new_descriptor() to create any number of BLACS
    descriptors sharing the same CPU layout.

    Most matrix operations require the involved matrices to all be on
    the same BlacsGrid.  Use a Redistributor to redistribute matrices
    from one BLACS grid to another if necessary.

    Parameters::

      * comm:  MPI communicator for CPUs of the BLACS grid or None.  A BLACS
        grid may use all or some of the CPUs of the communicator.
      * nprow:  Number of CPU rows.
      * npcol: Number of CPU columns.
      * order: 'R' or 'C', meaning rows or columns.  I'm not sure what this
        does, it probably interchanges the meaning of rows and columns. XXX

    Complicated stuff
    -----------------

    It may be useful to know that a BLACS grid is said to be active
    and will evaluate to True on any process where comm is not None
    *and* comm.rank < nprow * npcol.  Otherwise it is considered
    inactive and evaluates to False.  Ranks where a grid is inactive
    never do anything at all.

    BLACS identifies each grid by a unique ID number called the
    context (frequently abbreviated ConTxt).  Grids on inactive ranks
    have context -1."""
    def __init__(self, comm, nprow, npcol, order='R'):
        assert nprow > 0
        assert npcol > 0
        assert len(order) == 1
        assert order in 'CcRr'
        # set a default value for the context leads to fewer
        # if statements below
        self.context = INACTIVE

        self.comm = comm
        self.nprow = nprow
        self.npcol = npcol
        self.ncpus = nprow * npcol
        self.order = order

        # There are three cases to handle:
        # 1. Comm is None is inactive (default).
        # 2. Comm is a legitimate communicator
        # 3. DryRun
        if gpaw.dry_run:
            self.mycol = INACTIVE
            self.myrow = INACTIVE
            return

        if comm is not None:  # MPI task is part of the communicator
            if nprow * npcol > comm.size:
                raise ValueError(
                    'Impossible: %dx%d Blacs grid with %d CPUs'
                    % (nprow, npcol, comm.size))

            try:
                new = _gpaw.new_blacs_context
            except AttributeError as e:
                raise AttributeError(
                    'BLACS is unavailable.  '
                    'GPAW must be compiled with BLACS/ScaLAPACK, '
                    'and must run in MPI-enabled interpreter '
                    '(gpaw-python).  Original error: %s' % e)

            self.context = new(comm.get_c_object(), npcol, nprow, order)
            assert (self.context != INACTIVE) == (comm.rank < nprow * npcol)

        self.mycol, self.myrow = _gpaw.get_blacs_gridinfo(self.context,
                                                          nprow,
                                                          npcol)

    @property
    def coords(self):
        return self.myrow, self.mycol

    @property
    def shape(self):
        return self.nprow, self.npcol

    def coords2rank(self, row, col):
        return self.nprow * col + row

    def rank2coords(self, rank):
        col, row = divmod(rank, self.nprow)
        return row, col

    def new_descriptor(self, M, N, mb, nb, rsrc=0, csrc=0):
        """Create a new descriptor from this BLACS grid.

        See documentation for BlacsDescriptor.__init__."""
        return BlacsDescriptor(self, M, N, mb, nb, rsrc, csrc)

    def is_active(self):
        """Whether context is active on this rank."""
        return self.context != INACTIVE

    def __bool__(self):
        2 / 0

    def __str__(self):
        classname = self.__class__.__name__
        template = '%s[comm:size=%d,rank=%d; context=%d; %dx%d]'
        string = template % (classname, self.comm.size, self.comm.rank,
                             self.context, self.nprow, self.npcol)
        return string

    def __del__(self):
        if self.is_active():
            _gpaw.blacs_destroy(self.context)


class DryRunBlacsGrid00000000(BlacsGrid):
    def __init__(self, comm, nprow, npcol, order='R'):
        assert (isinstance(comm, SerialCommunicator) or
                isinstance(comm.comm, SerialCommunicator))
        # DryRunCommunicator is subclass

        if nprow * npcol > comm.size:
            raise ValueError('Impossible: %dx%d Blacs grid with %d CPUs'
                             % (nprow, npcol, comm.size))
        self.context = INACTIVE
        self.comm = comm
        self.nprow = nprow
        self.npcol = npcol
        self.ncpus = nprow * npcol
        self.mycol, self.myrow = INACTIVE, INACTIVE
        self.order = order


class BlacsDescriptor(MatrixDescriptor):
    """Class representing a 2D matrix distribution on a blacs grid.

    A BlacsDescriptor represents a particular shape and distribution
    of matrices.  A BlacsDescriptor has a global matrix shape and a
    rank-dependent local matrix shape.  The local shape is not
    necessarily equal on all ranks.

    A numpy array is said to be compatible with a BlacsDescriptor if,
    on all ranks, the shape of the numpy array is equal to the local
    shape of the BlacsDescriptor.  Compatible arrays can be created
    conveniently with the zeros() and empty() methods.

    An array with a global shape of M by N is distributed such that
    each process gets a number of distinct blocks of size mb by nb.
    The blocks on one process generally reside in very different areas
    of the matrix to improve load balance.

    The following chart describes how different ranks (there are 4
    ranks in this example, 0 through 3) divide the matrix into blocks.
    This is called 2D block cyclic distribution::

        +--+--+--+--+..+--+
        | 0| 1| 0| 1|..| 1|
        +--+--+--+--+..+--+
        | 2| 3| 2| 3|..| 3|
        +--+--+--+--+..+--+
        | 0| 1| 0| 1|..| 1|
        +--+--+--+--+..+--+
        | 2| 3| 2| 3|..| 3|
        +--+--+--+--+..+--+
        ...................
        ...................
        +--+--+--+--+..+--+
        | 2| 3| 2| 3|..| 3|
        +--+--+--+--+..+--+

    Also refer to:
    https://netlib.org/scalapack/slug/index.html

    Parameters:
     * blacsgrid: the BLACS grid of processors to distribute matrices.
     * M: global row count
     * N: global column count
     * mb: number of rows per block
     * nb: number of columns per block
     * rsrc: rank on which the first row is stored
     * csrc: rank on which the first column is stored

    Complicated stuff
    -----------------

    If there is trouble with matrix shapes, the below caveats are
    probably the reason.

    Depending on layout, a descriptor may have a local shape of zero
    by N or something similar.  If the row blocksize is 7, the global
    row count is 10, and the blacs grid contains 3 row processes: The
    first process will have 7 rows, the next will have 3, and the last
    will have 0.  The shapes in this case must still be correctly
    given to BLACS functions, which can be confusing.

    A blacs descriptor must also give the correct local leading
    dimension (lld), which is the local array size along the
    memory-contiguous direction in the matrix, and thus equal to the
    local column number, *except* when local shape is zero, but the
    implementation probably works.

    """
    def __init__(self, blacsgrid, M, N, mb, nb, rsrc=0, csrc=0):
        assert M > 0
        assert N > 0
        assert 1 <= mb
        assert 1 <= nb
        if mb > M:
            mb = M
        if nb > N:
            nb = N
        assert 0 <= rsrc < blacsgrid.nprow
        assert 0 <= csrc < blacsgrid.npcol

        self.blacsgrid = blacsgrid
        self.M = M  # global size 1
        self.N = N  # global size 2
        self.mb = mb  # block cyclic distr dim 1
        self.nb = nb  # and 2.  How many rows or columns are on this processor
        # more info:
        # http://www.netlib.org/scalapack/slug/node75.html
        self.rsrc = rsrc
        self.csrc = csrc

        if blacsgrid.is_active():
            locN, locM = _gpaw.get_blacs_local_shape(self.blacsgrid.context,
                                                     self.N, self.M,
                                                     self.nb, self.mb,
                                                     self.csrc, self.rsrc)
            # max 1 is nonsensical, but appears
            # to be required by PBLAS
            self.lld = max(1, locN)
        else:
            # ScaLAPACK has no requirements as to what these values on an
            # inactive blacsgrid should be. This seemed reasonable to me
            # at the time.
            locN, locM = 0, 0
            # This applies only to empty matrices.  Some functions will
            # complain about lld<1, so lld=1 will make those happy.
            self.lld = 1

        # locM, locN is not allowed to be negative. This will cause the
        # redistributor to fail. This could happen on active blacsgrid
        # which does not contain any piece of the distribute matrix.
        # This is why there is a final check on the value of locM, locN.
        MatrixDescriptor.__init__(self, max(0, locM), max(0, locN))

        # This is the definition of inactive descriptor; can occur
        # on an active or inactive blacs grid.
        self.active = locM > 0 and locN > 0

        self.bshape = (self.mb, self.nb)  # Shape of one block
        self.gshape = (M, N)  # Global shape of array

    def asarray(self):
        """Return a nine-element array representing this descriptor.

        In the C/Fortran code, a BLACS descriptor is represented by a
        special array of arcane nature.  The value of asarray() must
        generally be passed to BLACS functions in the C code."""
        arr = np.array([BLOCK_CYCLIC_2D, self.blacsgrid.context,
                        self.N, self.M, self.nb, self.mb, self.csrc, self.rsrc,
                        self.lld], np.intc)
        return arr

    def __repr__(self):
        classname = self.__class__.__name__
        template = '%s[context=%d, glob %s, block %s, lld %d, loc %s]'
        string = template % (classname, self.blacsgrid.context,
                             self.gshape,
                             self.bshape, self.lld, self.shape)
        return string

    def index2grid(self, row, col):
        """Get the BLACS grid coordinates storing global index (row, col)."""
        assert row < self.gshape[0], (row, col, self.gshape)
        assert col < self.gshape[1], (row, col, self.gshape)
        gridx = (row // self.bshape[0]) % self.blacsgrid.nprow
        gridy = (col // self.bshape[1]) % self.blacsgrid.npcol
        return gridx, gridy

    def index2rank(self, row, col):
        """Get the rank where global index (row, col) is stored."""
        return self.blacsgrid.coords2rank(*self.index2grid(row, col))

    def diagonalize_dc(self, H_nn, C_nn, eps_N, UL='L'):
        """See documentation in gpaw/utilities/blacs.py."""
        scalapack_diagonalize_dc(self, H_nn, C_nn, eps_N, UL)

    def diagonalize_ex(self, H_nn, C_nn, eps_N, UL='L', iu=None):
        """See documentation in gpaw/utilities/blacs.py."""
        scalapack_diagonalize_ex(self, H_nn, C_nn, eps_N, UL, iu=iu)

    def diagonalize_mr3(self, H_nn, C_nn, eps_N, UL='L', iu=None):
        """See documentation in gpaw/utilities/blacs.py."""
        scalapack_diagonalize_mr3(self, H_nn, C_nn, eps_N, UL, iu=iu)

    def general_diagonalize_dc(self, H_mm, S_mm, C_mm, eps_M,
                               UL='L'):
        """See documentation in gpaw/utilities/blacs.py."""
        scalapack_general_diagonalize_dc(self, H_mm, S_mm, C_mm, eps_M,
                                         UL)

    def general_diagonalize_ex(self, H_mm, S_mm, C_mm, eps_M,
                               UL='L', iu=None):
        """See documentation in gpaw/utilities/blacs.py."""
        scalapack_general_diagonalize_ex(self, H_mm, S_mm, C_mm, eps_M,
                                         UL, iu=iu)

    def general_diagonalize_mr3(self, H_mm, S_mm, C_mm, eps_M,
                                UL='L', iu=None):
        """See documentation in gpaw/utilities/blacs.py."""
        scalapack_general_diagonalize_mr3(self, H_mm, S_mm, C_mm, eps_M,
                                          UL, iu=iu)

    def inverse_cholesky(self, S_nn, UL='L'):
        """See documentation in gpaw/utilities/blacs.py."""
        scalapack_inverse_cholesky(self, S_nn, UL)

    def my_blocks(self, array_mn):
        """Yield the local blocks and their global index limits.

        Yields tuples of the form (Mstart, Mstop, Nstart, Nstop, block),
        for each locally stored block of the array.
        """
        if not self.check(array_mn):
            raise ValueError('Bad array shape (%s vs %s)' % (self,
                                                             array_mn.shape))

        grid = self.blacsgrid
        mb = self.mb
        nb = self.nb
        myrow = grid.myrow
        mycol = grid.mycol
        nprow = grid.nprow
        npcol = grid.npcol
        M, N = self.gshape

        Mmyblocks = -(-self.shape[0] // mb)
        Nmyblocks = -(-self.shape[1] // nb)
        for Mblock in range(Mmyblocks):
            for Nblock in range(Nmyblocks):
                myMstart = Mblock * mb
                myNstart = Nblock * nb
                Mstart = myrow * mb + Mblock * mb * nprow
                Nstart = mycol * mb + Nblock * nb * npcol
                Mstop = min(Mstart + mb, M)
                Nstop = min(Nstart + nb, N)
                block = array_mn[myMstart:myMstart + mb,
                                 myNstart:myNstart + nb]

                yield Mstart, Mstop, Nstart, Nstop, block

    def as_serial(self):
        return self.blacsgrid.new_descriptor(self.M, self.N, self.M, self.N)

    def redistribute(self, otherdesc, src_mn, dst_mn=None,
                     subM=None, subN=None, ia=0, ja=0, ib=0, jb=0, uplo='G'):
        if self.blacsgrid != otherdesc.blacsgrid:
            raise ValueError('Cannot redistribute to other BLACS grid.  '
                             'Requires using Redistributor class explicitly')
        if dst_mn is None:
            dst_mn = otherdesc.empty(dtype=src_mn.dtype)
        r = Redistributor(self.blacsgrid.comm, self, otherdesc)
        r.redistribute(src_mn, dst_mn, subM, subN, ia, ja, ib, jb, uplo)
        return dst_mn

    def collect_on_master(self, src_mn, dst_mn=None, uplo='G'):
        desc = self.as_serial()
        return self.redistribute(desc, src_mn, dst_mn, uplo=uplo)

    def distribute_from_master(self, src_mn, dst_mn=None, uplo='G'):
        desc = self.as_serial()
        return desc.redistribute(self, src_mn, dst_mn, uplo=uplo)


class Redistributor:
    """Class for redistributing BLACS matrices on different contexts."""
    def __init__(self, supercomm, srcdescriptor, dstdescriptor):
        """Create redistributor.

        Source and destination descriptors may reside on different
        BLACS grids, but the descriptors should describe arrays with
        the same number of elements.

        The communicators of the BLACS grid of srcdescriptor as well
        as that of dstdescriptor *must* both be subcommunicators of
        supercomm.

        Allowed values of UPLO are: G for general matrix, U for upper
        triangular and L for lower triangular. The latter two are useful
        for symmetric matrices."""
        self.supercomm = supercomm
        self.supercomm_bg = BlacsGrid(self.supercomm, self.supercomm.size, 1)
        self.srcdescriptor = srcdescriptor
        self.dstdescriptor = dstdescriptor

    def redistribute(self, src_mn, dst_mn=None,
                     subM=None, subN=None,
                     ia=0, ja=0, ib=0, jb=0, uplo='G'):
        """Redistribute src_mn into dst_mn.

        src_mn and dst_mn must be compatible with source and
        destination descriptors of this redistributor.

        If subM and subN are given, distribute only a subM by subN
        submatrix.

        If any ia, ja, ib and jb are given, they denote the global
        index (i, j) of the origin of the submatrix inside the source
        and destination (a, b) matrices."""

        srcdescriptor = self.srcdescriptor
        dstdescriptor = self.dstdescriptor
        dtype = src_mn.dtype

        if dst_mn is None:
            dst_mn = dstdescriptor.empty(dtype=dtype)

        # self.supercomm must be a supercommunicator of the communicators
        # corresponding to the context of srcmatrix as well as dstmatrix.
        # We should verify this somehow.
        srcdescriptor = self.srcdescriptor
        dstdescriptor = self.dstdescriptor

        dtype = src_mn.dtype
        if dst_mn is None:
            dst_mn = dstdescriptor.zeros(dtype=dtype)

        assert dtype == dst_mn.dtype
        assert dtype == float or dtype == complex

        # Check to make sure the submatrix of the source
        # matrix will fit into the destination matrix
        # plus standard BLACS matrix checks.
        srcdescriptor.checkassert(src_mn)
        dstdescriptor.checkassert(dst_mn)

        if subM is None:
            subM = srcdescriptor.gshape[0]
        if subN is None:
            subN = srcdescriptor.gshape[1]

        assert srcdescriptor.gshape[0] >= subM
        assert srcdescriptor.gshape[1] >= subN
        assert dstdescriptor.gshape[0] >= subM
        assert dstdescriptor.gshape[1] >= subN

        # Switch to Fortran conventions
        uplo = {'U': 'L', 'L': 'U', 'G': 'G'}[uplo]
        _gpaw.scalapack_redist(srcdescriptor.asarray(),
                               dstdescriptor.asarray(),
                               src_mn, dst_mn,
                               subN, subM,
                               ja + 1, ia + 1, jb + 1, ib + 1,  # 1-indexing
                               self.supercomm_bg.context, uplo)
        return dst_mn


def parallelprint(comm, obj):
    import sys
    for a in range(comm.size):
        if a == comm.rank:
            print('rank=%d' % a)
            print(obj)
            print()
            sys.stdout.flush()
        comm.barrier()



import time
import pytest
import numpy as np

from ase.build import bulk
from ase.parallel import parprint

from gpaw import GPAW, PW
from gpaw.test import findpeak
from gpaw.bztools import find_high_symmetry_monkhorst_pack
from gpaw.response.df import DielectricFunction, read_response_function
from gpaw.mpi import size, world


# Affected by https://gitlab.com/gpaw/gpaw/-/issues/840
# We are disabling assertions below as necessary, should be reenabled
# after fixing 840.
@pytest.mark.response
def test_response_aluminum_EELS_RPA(in_tmp_dir):
    assert size <= 4**3

    # Ground state calculation

    t1 = time.time()

    a = 4.043
    atoms = bulk('Al', 'fcc', a=a)
    atoms.center()
    calc = GPAW(mode=PW(200),
                nbands=4,
                kpts=(4, 4, 4),
                parallel={'band': 1},
                xc='LDA')

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('Al_gs')

    # Generate grid compatible with tetrahedron integration
    kpts = find_high_symmetry_monkhorst_pack('Al_gs', 2.0)

    # Calculate the wave functions on the new kpts grid
    calc = GPAW('Al_gs').fixed_density(kpts=kpts)
    calc.write('Al', 'all')

    t2 = time.time()

    # Excited state calculation
    q0_c = np.array([0., 0., 0.])
    q1_c = np.array([1 / 4., 0., 0.])
    w_w = np.linspace(0, 24, 241)

    # Calculate the eels spectrum using point integration at both q-points
    df1 = DielectricFunction(calc='Al', frequencies=w_w, eta=0.2, ecut=50,
                             hilbert=False, rate=0.2)
    df1.get_eels_spectrum(xc='RPA', filename='EELS_Al-PI_q0', q_c=q0_c)
    df1.get_eels_spectrum(xc='RPA', filename='EELS_Al-PI_q1', q_c=q1_c)

    t3 = time.time()

    # Calculate the eels spectrum using tetrahedron integration at q=0
    # NB: We skip the finite q-point, because the underlying symmetry
    # exploration runs excruciatingly slowly at finite q...
    df2 = DielectricFunction(calc='Al', eta=0.2, ecut=50,
                             integrationmode='tetrahedron integration',
                             hilbert=True, rate=0.2)
    df2.get_eels_spectrum(xc='RPA', filename='EELS_Al-TI_q0', q_c=q0_c)

    t4 = time.time()

    parprint('')
    parprint('For ground  state calc, it took', (t2 - t1) / 60, 'minutes')
    parprint('For PI excited state calc, it took', (t3 - t2) / 60, 'minutes')
    parprint('For TI excited state calc, it took', (t4 - t3) / 60, 'minutes')

    world.barrier()
    omegaP0_w, eels0P0_w, eelsP0_w = read_response_function('EELS_Al-PI_q0')
    omegaP1_w, eels0P1_w, eelsP1_w = read_response_function('EELS_Al-PI_q1')
    omegaT0_w, eels0T0_w, eelsT0_w = read_response_function('EELS_Al-TI_q0')

    # calculate the 1 & 2 wpeak and ipeak values for tetra and point int.
    wpeak1P0, Ipeak1P0 = findpeak(omegaP0_w, eels0P0_w)
    wpeak2P0, Ipeak2P0 = findpeak(omegaP0_w, eelsP0_w)
    wpeak1P1, Ipeak1P1 = findpeak(omegaP1_w, eels0P1_w)
    wpeak2P1, Ipeak2P1 = findpeak(omegaP1_w, eelsP1_w)
    wpeak1T0, Ipeak1T0 = findpeak(omegaT0_w, eels0T0_w)
    wpeak2T0, Ipeak2T0 = findpeak(omegaT0_w, eelsT0_w)

    # import matplotlib.pyplot as plt
    # plt.plot(omegaP0_w, eelsP0_w)
    # plt.plot(omegaT0_w, eelsT0_w)
    # plt.show()

    # tetra and point integrators should produce similar results; however,
    # Al converges very slowly w.r.t. kpts so we just make sure the
    # values don't change and tests consistency elsewhere
    assert wpeak1P0 == pytest.approx(15.7111, abs=0.02)
    assert wpeak2P0 == pytest.approx(15.7096, abs=0.02)
    assert wpeak1P1 == pytest.approx(15.8402, abs=0.02)
    assert wpeak2P1 == pytest.approx(15.8645, abs=0.02)
    # assert wpeak1T0 == pytest.approx(20.2119, abs=0.02)  # XXX #840
    # assert wpeak2T0 == pytest.approx(20.2179, abs=0.02)  # XXX #840

    assert Ipeak1P0 == pytest.approx(29.40, abs=1.)
    assert Ipeak2P0 == pytest.approx(27.70, abs=1.)
    assert Ipeak1P1 == pytest.approx(28.39, abs=1.)
    assert Ipeak2P1 == pytest.approx(26.89, abs=1.)
    # assert Ipeak1T0 == pytest.approx(46.24, abs=1.)  # XXX #840
    # assert Ipeak2T0 == pytest.approx(44.27, abs=1.)  # XXX #840



import numpy as np


def default_rng(seed):
    return RNG(np.random.default_rng(seed))


class RNG:
    def __init__(self, rng):
        self.rng = rng

    def random(self, shape, out):
        self.rng.random(shape, out=out._data)



import _gpaw
from gpaw.fd_operators import FDOperator
import numpy as np


class WeightedFDOperator(FDOperator):
    def __init__(self, operators):
        """Compound operator A with nonlocal weights.

        A = Sum_i weights_i * operators_i

        Arguments:
        operators -- List of FDOperators.

        The weights have to be set using the set_weights method.
        A is build from operators which are then discarded.
        """
        self.gd = operators[0].gd
        self.shape = tuple(self.gd.n_c)
        self.coef_ps = []
        self.offset_ps = []
        for op in operators:
            assert self.gd == op.gd
            assert operators[0].dtype == op.dtype
            assert not hasattr(op, 'nweights')
            if 0 in op.offset_p:
                assert op.offset_p[0] == 0
                self.offset_ps.append(
                    np.ascontiguousarray(op.offset_p.copy())
                )
                self.coef_ps.append(
                    np.ascontiguousarray(op.coef_p.copy())
                )
            else:
                self.offset_ps.append(
                    np.ascontiguousarray(np.hstack(([0], op.offset_p))))
                self.coef_ps.append(
                    np.ascontiguousarray(np.hstack(([.0], op.coef_p))))
            assert self.offset_ps[-1][0] == 0
            assert len(self.coef_ps[-1]) == len(self.offset_ps[-1])
            assert self.offset_ps[-1].flags.c_contiguous
            assert self.coef_ps[-1].flags.c_contiguous
            assert self.offset_ps[-1].dtype == int
            assert self.coef_ps[-1].dtype == float

        self.nweights = len(operators)
        self.cfd = all([op.cfd for op in operators])
        self.mp = max([op.mp for op in operators])
        self.dtype = operators[0].dtype
        if self.gd.comm.size > 1:
            self.comm = self.gd.comm.get_c_object()
        else:
            self.comm = None
        self.npoints = max([op.npoints for op in operators])
        self.weights = None
        self.operator = None
        self.description = 'Weighted Finite Difference Operator\n      '
        self.description += '\n      '.join([op.description
                                             for op in operators])

        self.xp = np

    def set_weights(self, weights):
        """Set the operator weights.

        weights   -- List of numpy arrays sized gd.n_c
                     (weights are not copied).
        """
        assert len(weights) == self.nweights
        for weight in weights:
            assert weight.shape == self.shape
            assert weight.dtype == float
            assert weight.flags.c_contiguous
        self.weights = weights
        self.operator = _gpaw.WOperator(
            self.nweights, self.weights,
            self.coef_ps, self.offset_ps, self.gd.n_c, self.mp,
            self.gd.neighbor_cd, self.dtype == float, self.comm, self.cfd
        )



import scipy.fft as fft
import gpaw.gpu.cpupy as cp


def ifftn(array_Q,
          shape,
          norm=None,
          overwrite_x=False):
    return cp.ndarray(fft.ifftn(array_Q._data,
                                shape,
                                norm=norm,
                                overwrite_x=overwrite_x))


def fftn(array_Q,
         shape=None,
         norm=None,
         overwrite_x=False):
    return cp.ndarray(fft.fftn(array_Q._data,
                               shape,
                               norm=norm,
                               overwrite_x=overwrite_x))


def rfftn(array_Q,
          shape=None,
          norm=None,
          overwrite_x=False):
    return cp.ndarray(fft.rfftn(array_Q._data,
                                shape,
                                norm=norm,
                                overwrite_x=overwrite_x))


def irfftn(array_Q,
           shape=None,
           norm=None,
           overwrite_x=False):
    return cp.ndarray(fft.irfftn(array_Q._data,
                                 shape,
                                 norm=norm,
                                 overwrite_x=overwrite_x))
